"""Extract the main fields from an SQL database into JSON format."""

import pathlib
import sqlite3

import click
import orjson
import tqdm
from context_verbose import Printer
from cutcutcodec.utils import mround

from mendevi.database.meta import merge_extractors

from .parse import parse_videos_database


def _round(value: object) -> object:
    """Alias to cutcutcodec mround for float only."""
    if isinstance(value, float):
        return mround(value)
    return value


def _add_key_key_val(dico: dict, key1: str, key2: str, value: object) -> None:
    """Ensure dico[key1][key2] = [value, ...]."""
    dico[key1] = dico.get(key1, {})
    dico[key1][key2] = dico[key1].get(key2, [])
    dico[key1][key2].append(value)


@click.command()
@click.argument("database", type=click.Path())
@click.option("-o", "--output", type=click.Path(), help="The output json database path.")
def main(database: str, **kwargs: dict) -> None:
    """Extract the main fields from an SQL database into JSON format.

    \b
    Parameters
    ----------
    database : pathlike
        The source SQL database to be converted.
    **kwargs: dict
        Please refer to the detailed arguments below.
    output : pathlike, optional
        The destination json database path.
        By default, the file is created in the same folder than the SQL database.

    """
    # parse args
    with Printer("Parse configuration...") as prt:
        _, database = parse_videos_database(prt, (), database)
        if (output := kwargs.get("output")) is None:
            output = database.with_suffix(".json")
        else:
            output = pathlib.Path(output).expanduser()
            if output.is_dir():
                output = output / f"{database.stem}.json"
        prt.print(f"json file : {output}")

    # read database
    content = {}
    context = [
        "bitrate",
        "effort",
        "encoder",
        "eotf",
        "height",
        "lpips",
        "mode",
        "primaries",
        "profile",
        "psnr",
        "quality",
        "range",
        "ref_name",
        "ssim",
        "threads",
        "video_duration",
        "vmaf",
        "width",
    ]
    with Printer("Read SQL database...", color="cyan") as prt:
        prt.print("compile the line extractor")
        _atom_names, line_extractor = merge_extractors(
            {
                "video_hash",
                # context
                *context,
                # activity
                "hostname",
                "act_duration",
                "cores",
                "power",
                "temp",
            },
            return_callable=True,
        )
        with sqlite3.connect(database) as conn:
            conn.row_factory = sqlite3.Row
            prt.print("read 'encode' and 'metric' extractor")
            # from mendevi.database.meta import get_extractor
            # from mendevi.database.extract import SqlLinker
            # select = {s for lbl in atom_names for s in get_extractor(lbl).func.select}
            # for query in SqlLinker(*select).sql:
            #     print(query)
            count = conn.execute("SELECT COUNT(*) FROM t_enc_encode").fetchone()["COUNT(*)"]
            for row_ in tqdm.tqdm(
                conn.execute(
                    """
                    SELECT
                        t_act_activity.act_duration,
                        t_act_activity.act_ps_core,
                        t_act_activity.act_ps_dt,
                        t_act_activity.act_ps_temp,
                        t_act_activity.act_rapl_dt,
                        t_act_activity.act_rapl_power,
                        t_act_activity.act_wattmeter_dt,
                        t_act_activity.act_wattmeter_power,
                        t_dst_video.vid_duration,
                        t_dst_video.vid_eotf,
                        t_dst_video.vid_gamut,
                        t_dst_video.vid_height,
                        t_dst_video.vid_id,
                        t_dst_video.vid_range,
                        t_dst_video.vid_size,
                        t_dst_video.vid_width,
                        t_enc_encode.enc_effort,
                        t_enc_encode.enc_encoder,
                        t_enc_encode.enc_mode,
                        t_enc_encode.enc_quality,
                        t_enc_encode.enc_threads,
                        t_env_environment.env_hostname,
                        t_met_metric.met_lpips_alex,
                        t_met_metric.met_lpips_vgg,
                        t_met_metric.met_psnr,
                        t_met_metric.met_ssim,
                        t_met_metric.met_vmaf,
                        t_ref_video.vid_name AS ref_vid_name
                    FROM t_enc_encode
                    JOIN t_vid_video AS t_ref_video
                        ON t_enc_encode.enc_src_vid_id = t_ref_video.vid_id
                    JOIN t_vid_video AS t_dst_video
                        ON t_enc_encode.enc_dst_vid_id = t_dst_video.vid_id
                    LEFT JOIN t_met_metric
                        ON t_enc_encode.enc_dst_vid_id = t_met_metric.met_dis_vid_id
                        AND t_enc_encode.enc_src_vid_id = t_met_metric.met_ref_vid_id
                    JOIN t_act_activity
                        ON t_enc_encode.enc_act_id = t_act_activity.act_id
                    JOIN t_env_environment
                        ON t_enc_encode.enc_env_id = t_env_environment.env_id
                    """,
                ),
                dynamic_ncols=True,
                leave=False,
                smoothing=1e-6,
                total=count,
                unit="vid",
            ):
                row = line_extractor(dict(row_))
                vid_id = row.pop("video_hash")
                content[vid_id] = content.get(vid_id, {"id": vid_id})
                content[vid_id] |= {k: _round(row[k]) for k in context if row[k] is not None}
                _add_key_key_val(
                    content[vid_id],
                    "encode_duration",
                    row["hostname"],
                    _round(row["act_duration"]),
                )
                _add_key_key_val(
                    content[vid_id], "encode_cores", row["hostname"], _round(row["cores"]),
                )
                _add_key_key_val(
                    content[vid_id], "encode_power", row["hostname"], _round(row["power"]),
                )
                _add_key_key_val(
                    content[vid_id], "encode_temp", row["hostname"], _round(row["temp"]),
                )
            prt.print("read 'decode' and 'metric' extractor")
            count = conn.execute("SELECT COUNT(*) FROM t_dec_decode").fetchone()["COUNT(*)"]
            for row_ in tqdm.tqdm(
                conn.execute(
                    """
                    SELECT
                        t_act_activity.act_duration,
                        t_act_activity.act_ps_core,
                        t_act_activity.act_ps_dt,
                        t_act_activity.act_ps_temp,
                        t_act_activity.act_rapl_dt,
                        t_act_activity.act_rapl_power,
                        t_act_activity.act_wattmeter_dt,
                        t_act_activity.act_wattmeter_power,
                        t_dst_video.vid_duration,
                        t_dst_video.vid_eotf,
                        t_dst_video.vid_gamut,
                        t_dst_video.vid_height,
                        t_dst_video.vid_id,
                        NULL AS ref_vid_name,  -- bulshit
                        t_dst_video.vid_range,
                        t_dst_video.vid_size,
                        t_dst_video.vid_width,
                        NULL AS enc_effort,  -- bulshit
                        NULL AS enc_encoder,  -- bulshit
                        NULL AS enc_mode,  -- bulshit
                        NULL AS enc_quality,  -- bulshit
                        NULL AS enc_threads,  -- bulshit
                        t_env_environment.env_hostname,
                        t_met_metric.met_lpips_alex,
                        t_met_metric.met_lpips_vgg,
                        t_met_metric.met_psnr,
                        t_met_metric.met_ssim,
                        t_met_metric.met_vmaf
                    FROM t_dec_decode
                    JOIN t_act_activity
                        ON t_dec_decode.dec_act_id = t_act_activity.act_id
                    LEFT JOIN t_met_metric
                        ON t_dec_decode.dec_vid_id = t_met_metric.met_dis_vid_id
                    JOIN t_vid_video AS t_dst_video
                        ON t_dec_decode.dec_vid_id = t_dst_video.vid_id
                    JOIN t_env_environment
                        ON t_dec_decode.dec_env_id = t_env_environment.env_id
                    """,
                ),
                dynamic_ncols=True,
                leave=False,
                smoothing=1e-6,
                total=count,
                unit="vid",
            ):
                row = line_extractor(dict(row_))
                row["name"] = None
                vid_id = row.pop("video_hash")
                content[vid_id] = content.get(vid_id, {"id": vid_id})
                content[vid_id] |= {k: _round(row[k]) for k in context if row[k] is not None}
                _add_key_key_val(
                    content[vid_id],
                    "decode_duration",
                    row["hostname"],
                    _round(row["act_duration"]),
                )
                _add_key_key_val(
                    content[vid_id], "decode_cores", row["hostname"], _round(row["cores"]),
                )
                _add_key_key_val(
                    content[vid_id], "decode_power", row["hostname"], _round(row["power"]),
                )
                _add_key_key_val(
                    content[vid_id], "decode_temp", row["hostname"], _round(row["temp"]),
                )
            prt.print_time()

    # write json
    content = [content[k] for k in sorted(content)]
    with Printer("Write JSON database...", color="cyan") as prt:
        content = orjson.dumps(content, option=orjson.OPT_INDENT_2|orjson.OPT_SORT_KEYS)
        with output.open("wb") as raw:
            raw.write(content)
        prt.print_time()



"""Help to excecute a script on a node of grid'5000."""

import getpass
import hashlib
import json
import logging
import pathlib
import platform
import re
import subprocess
import time

import click
import orjson
import requests
from context_verbose import Printer

GOOD_REQ_STATUS = 200
SLEEP_FINAL = 600.0  # final sleeping time
SLEEP_INIT = 10.0  # initial sleeping time
SLEEP_TRANSITION = 3600.0  # the transition time between the init and final sleep


def _parse_args(prt: Printer, kwargs: dict) -> None:
    """Verification of the arguments."""
    # env_name
    kwargs["env_name"] = kwargs.get("env_name", "")
    if not kwargs["env_name"]:
        match len(env_name := guess_envname()):
            case 0:
                msg = (
                    "No environment .yaml images were found.\n"
                    "You must therefore specify it with the argument --env-name.\n"
                    'See: "https://www.grid5000.fr/w/Getting_Started'
                    '#Deploying_your_nodes_to_get_root_access_and_'
                    'create_your_own_experimental_environment".'
                )
                raise OSError(msg)
            case 1:
                kwargs["env_name"] = env_name[0]
            case _:
                logging.getLogger(__name__).warning(
                    "serveral env name were founded %s, please specify --env-name",
                    env_name,
                )
    else:
        assert isinstance(kwargs["env_name"], str), kwargs["env_name"].__class__.__name__
    prt.print(f"env_name: {kwargs['env_name']}")

    # walltime
    kwargs["walltime"] = kwargs.get("walltime") or "24:00:00"
    assert isinstance(kwargs["walltime"], str), kwargs["walltime"].__class__.__name__
    prt.print(f"walltime: {kwargs['walltime']}")

    # roles
    kwargs["roles"] = kwargs.get("roles", []) or guess_roles()
    assert isinstance(kwargs["roles"], tuple | list), kwargs["roles"].__class__.__name__
    kwargs["roles"] = list(kwargs["roles"])
    assert all(isinstance(r, str) for r in kwargs["roles"]), kwargs["roles"]
    prt.print(f"roles   : {kwargs['roles']}")

    # server
    kwargs["server"] = kwargs.get("server", ())
    assert isinstance(kwargs["server"], tuple), kwargs["server"].__class__.__name__
    assert all(isinstance(s, str) for s in kwargs["server"]), kwargs["server"]
    if kwargs["server"]:
        prt.print(f"servers : {', '.join(kwargs['server'])}")

    # cluster
    assert "cluster" in kwargs, kwargs
    assert isinstance(kwargs["cluster"], str), kwargs["cluster"].__class__.__name__
    prt.print(f"cluster : {kwargs['cluster']}")

    # nodes
    kwargs["nodes"] = kwargs.get("nodes", 1)
    assert isinstance(kwargs["nodes"], int), kwargs["nodes"].__class__.__name__
    prt.print(f"nodes   : {kwargs['nodes']}")


def guess_envname() -> list[str]:
    """Try to find the envname.

    Returns
    -------
    envnames : list[str]
        The candidate envnames, as a list of urls formated like:
        https://api.grid5000.fr/stable/sites/{{site}}/public/{{username}}/{{file}}

    """
    envnames: list[str] = []
    root = pathlib.Path.home() / "public"
    for path in root.rglob("*.yaml"):
        with path.open("r", encoding="utf-8") as file:
            content = file.read()
        if re.search(r"file:\s*https?://[\w.]+\.grid5000\.fr", content):
            envnames.append(
                "https://api.grid5000.fr/stable"
                f"/sites/{platform.node()[1:]}"  # "frennes", "fgrenoble", "ftoulouse" ...
                f"/public/{getpass.getuser()}"  # os.getlogin failed sometimes
                f"/{path.relative_to(root)}",
            )
    return envnames


def guess_roles() -> list[str]:
    """Return the roles list.

    Returns
    -------
    roles : list[str]
        The name of the group to which one belongs

    """
    roles = re.findall(
        r"\S+",
        subprocess.run(["id", "-Gn"], capture_output=True, check=True).stdout.decode("utf-8"),
    )
    return [r for r in roles if r != "g5k-users"]


def guess_servers() -> list[str]:
    """Return the available nodes address with a wattmeter."""
    # see oarnodes -l

    # get full node list
    try:
        servers = subprocess.run(["oarnodes", "--json"], capture_output=True, check=True)
    except FileNotFoundError:
        return []
    servers = orjson.loads(servers.stdout)

    # select with wattmeter
    servers = {
        ref: desc for ref, desc in servers.items()
        if (
            desc["wattmeter"] in {True, "yes", "YES", "ok", "OK"}
            and desc["deploy"]
        )
    }
    return list(servers)


def guess_storage() -> list[str]:
    """Return the accessible paths under "/srv/storage".

    If they have not been mounted for a while, NFS disks may be hidden with a simple ls command.
    This function attempts to predict the volume name in order to increase the chances of success.

    The rules are defined here: https://www.grid5000.fr/w/Group_Storage
    """
    # guessed with ls
    if not (root := pathlib.Path("/srv/storage")).exists():
        return []
    candidates = set(root.iterdir())

    # try to get server_name
    url = "https://www.grid5000.fr/w/Group_Storage"
    req = requests.get(url, timeout=60)
    if req.status_code != GOOD_REQ_STATUS:
        logging.getLogger(__name__).warning("the request %s failed", req)
        servers = []
    elif not (req := req.text):
        logging.getLogger(__name__).warning("the request %s gives an empty result", req)
        servers = []
    else:
        servers = re.findall(r"storage\d+\.\w+\.grid5000\.fr", req)

    # try to get storage_name
    prefixes = guess_roles()
    prefixes.extend(n for s in prefixes.copy() for n in s.split("-"))

    # merge candidates together with a cartesian product
    candidates |= {root / f"{p}@{s}" for s in set(servers) for p in set(prefixes)}

    # test if exists and permision (.exists() doesn't work with this nfs)
    storages: list[str] = []
    for path in candidates:
        try:
            with (path / "tmp_empty_mendevi").open("wb"):
                pass
        except (FileNotFoundError, PermissionError):
            continue
        (path / "tmp_empty_mendevi").unlink()
        storages.append(str(path))
    return storages


def is_g5k_frontend(*, fail: bool = True) -> bool:
    """Test if we are on the grid'5000 frontend."""
    assert isinstance(fail, bool), fail.__class__.__name__
    if pathlib.Path("/etc/grid5000").exists():
        return True
    if not fail:
        return False
    msg = (
        "You are not on the Grid'5000 frontend!\n"
        "The command 'mendevi g5k ...' must be executed exclusively from the frontend.\n"
        'See: "https://www.grid5000.fr/w/Getting_Started#Connect_to_a_Grid\'5000_access_machine"'
    )
    raise OSError(
        msg,
    )


def set_ansible_cfg() -> None:
    """Create and fill the file ~/.ansible.cfg."""
    path = pathlib.Path.home() / ".ansible.cfg"
    if not path.exists():
        with path.open("w", encoding="utf-8") as file:
            file.write("[defaults]\n")
            file.write("interpreter_python = /root/.pyenv/shims/python\n")


@click.command()
@click.argument("cmd", type=str)
@click.option(
    "-e", "--env-name",
    type=str,
    help="The kadeploy3 environment to use.",
)
@click.option(
    "-w", "--walltime",
    type=str,
    help="Job duration.",
)
@click.option(
    "-s", "server",
    type=str,
    multiple=True,
    help="The specific node name.",
)
@click.option(
    "-c", "--cluster",
    type=str,
    help="The generique cluster name to use.",
)
@click.option(
    "-r", "roles",
    type=str,
    multiple=True,
    help="The specific role to use.",
)
@click.option(
    "-n", "--nodes",
    type=int,
    default=1,
    help="The number of nodes to be deployed.",
)
def main(cmd: str, **kwargs: dict) -> None:
    """Deploy a grid'5000 node and run a mendevi script inside it.

    \b
    Parameters
    ----------
    **kwargs: dict
        Please refer to the detailed arguments below.
    cmd : str
        The mendevi command to be executed on the deployed node.
        When a path begins with ./, this part is automatically replaced by the shared folder file.
    env_name : str, optional
        The kadeploy3 yaml path file environment to use.
        By default, the first environment found is used.
    walltime : str, default=24:00:00
        The maximum reservation time for the node.
        If the script finishes before this time, the resources are released before the expiry date.
    roles : tuple[str], default=autodetect
        The role to used, by default, the list is autodetected with :py:func:`guess_roles`.
    server : tuple[str], optional
        The specific node names (ex: ("paradoxe-1.rennes.grid5000.fr",)).
    cluster : str
        The name of the cluster to used (ex: "paradoxe").
    nodes : int, default=1
        The number of nodes to deploy (incompatible with ``server``).

    """
    is_g5k_frontend()
    set_ansible_cfg()
    with Printer("Parse configuration...") as prt:
        assert isinstance(cmd, str), cmd.__class__.__name__
        if " ./" in cmd and len(data := guess_storage()) == 1:
            cmd = re.split(r"\s+\./", cmd)
            cmd = f"{cmd[0]}{''.join(f' {data[0]}/{c}' for c in cmd[1:])}"
        prt.print(f"cmd     : {cmd}")
        _parse_args(prt, kwargs)

    # run on the grid
    import enoslib  # noqa: PLC0415
    enoslib.init_logging()
    conf = (
        # doc: https://discovery.gitlabpages.inria.fr/enoslib/apidoc/infra.html#g5k-schema
        enoslib.G5kConf.from_settings(
            env_name=kwargs["env_name"],
            job_name=f"mendevi {hashlib.md5(cmd.encode('utf-8')).hexdigest()}",
            job_type=["deploy", "exotic"],
            walltime=kwargs["walltime"],
            monitor="wattmetre_power_watt",
            # queue="production",  # "default", "testing", "besteffort"
        )
        .add_machine(
            roles=kwargs["roles"],
            cluster=kwargs["cluster"],
            nodes=kwargs["nodes"],
            **({"servers": list(kwargs["server"])} if kwargs["server"] else {}),
        )
    )
    sleep = SLEEP_INIT
    warnings: set[str] = set()  # the displayed warnings
    with enoslib.G5k(conf) as (roles, _):  # get and release all grid5000 resources
        main_results = enoslib.run_command(cmd, roles=roles, background=True)
        log_files = {result.results_file for result in main_results}
        while log_files:
            # waiting a moment
            # at the beginning, we wait SLEEP_INIT, and at the end, we wait SLEEP_FINAL
            # it takes n steps, with n = math.log(SLEEP_FINAL/SLEEP_INIT) / math.log(alpha)
            # with alpha = 1.0 + (SLEEP_FINAL - SLEEP_INIT) / SLEEP_TRANSITION
            # the full transition time take SLEEP_TRANSITION
            time.sleep(sleep)
            sleep *= 1.0 + (SLEEP_FINAL - SLEEP_INIT) / SLEEP_TRANSITION
            sleep = min(sleep, SLEEP_FINAL)
            for log_file, res in [
                (log_file_, res_)
                for log_file_ in log_files.copy()
                for res_ in enoslib.run_command(
                    (
                        f'if [ -f "{log_file_}" ]; '
                        f'then cat "{log_file_}"; '
                        'else echo "<no-log-yet>"; fi'
                    ),
                    roles=roles,
                    on_error_continue=True,
                )
            ]:
                # display warnings
                for msg in res.payload.get("warnings", []):
                    if msg not in warnings:
                        logging.getLogger(__name__).warning(msg)
                        warnings.add(msg)

                # skip when file log is not yet created
                try:
                    if res.payload["stdout"] == "<no-log-yet>":
                        continue
                except KeyError:
                    logging.getLogger(__name__).exception("stop communication with %s", log_files)
                    log_files.remove(log_file)
                    continue

                # get job result
                try:
                    res_ = json.loads(res.payload["stdout"])  # orjson failed
                except json.JSONDecodeError:
                    logging.getLogger(__name__).exception(res.payload)
                    raise
                if res_.get("changed", False) or "rc" in res_:
                    print(res_.get("stdout", ""))
                    logging.getLogger(__name__).error(res_.get("stderr", ""))
                    log_files.remove(log_file)



"""Test the ffmpeg encoding command on a stupid example."""

import itertools
import shlex
import subprocess

from mendevi.ffmpeg_cmd import (
    _encode_av1_nvenc,
    _encode_av1_vaapi,
    _encode_h264_nvenc,
    _encode_h264_vaapi,
    _encode_hevc_nvenc,
    _encode_hevc_vaapi,
    _encode_libaomav1,
    _encode_libopenh264,
    _encode_librav1e,
    _encode_libsvtav1,
    _encode_libvpx_vp9,
    _encode_libx264,
    _encode_libx265,
    _encode_vp9_vaapi,
    _encode_vvc,
)

RESOLUTION = [256, 256]
FFMPEG_CMD_READ = [
    "ffmpeg", "-hide_banner",
    "-f", "rawvideo",
    "-s", f"{RESOLUTION[1]}x{RESOLUTION[0]}",
    "-pix_fmt", "yuv420p",
    "-r", "24",
    "-i", "/dev/urandom",
    "-to", "0.25",
]


def encode(encoder: str, cmd_creator: callable) -> None:
    """General encoding test."""
    for mode, quality, effort, pix_fmt in itertools.product(
        ["vbr", "cbr"], [0.33, 0.66], ["fast", "medium", "slow"], ["yuv420p", "yuv420p10le"],
    ):
        cmd_middle = cmd_creator(
            effort=effort,
            encoder=encoder,
            mode=mode,
            pix_fmt=pix_fmt,
            quality=quality,
            resolution=RESOLUTION,
            threads=8,
        )
        cmd = [*FFMPEG_CMD_READ, *cmd_middle, "-f", "null", "-"]
        try:
            subprocess.run(cmd, check=True)
        except subprocess.CalledProcessError as err:
            msg = " ".join(map(shlex.quote, cmd))
            raise RuntimeError(msg) from err


def test_encode_av1_nvenc() -> None:
    """Perform the ffmpeg encoding test on av1_nvenc."""
    encode("av1_nvenc", _encode_av1_nvenc)


def test_encode_av1_vaapi() -> None:
    """Perform the ffmpeg encoding test on av1_vaapi."""
    encode("av1_vaapi", _encode_av1_vaapi)


def test_encode_h264_nvenc() -> None:
    """Perform the ffmpeg encoding test on h264_nvenc."""
    encode("h264_nvenc", _encode_h264_nvenc)


def test_encode_h264_vaapi() -> None:
    """Perform the ffmpeg encoding test on h264_vaapi."""
    encode("h264_vaapi", _encode_h264_vaapi)


def test_encode_hevc_nvenc() -> None:
    """Perform the ffmpeg encoding test on hevc_nvenc."""
    encode("hevc_nvenc", _encode_hevc_nvenc)


def test_encode_hevc_vaapi() -> None:
    """Perform the ffmpeg encoding test on hevc_vaapi."""
    encode("hevc_vaapi", _encode_hevc_vaapi)


def test_encode_libaomav1() -> None:
    """Perform the ffmpeg encoding test on libaom-av1."""
    encode("libaom-av1", _encode_libaomav1)


def test_encode_libopenh264() -> None:
    """Perform the ffmpeg encoding test on libopenh264."""
    encode("libopenh264", _encode_libopenh264)


def test_encode_librav1e() -> None:
    """Perform the ffmpeg encoding test on librav1e."""
    encode("librav1e", _encode_librav1e)


def test_encode_libsvtav1() -> None:
    """Perform the ffmpeg encoding test on libsvtav1."""
    encode("libsvtav1", _encode_libsvtav1)


def test_encode_libvpx_vp9() -> None:
    """Perform the ffmpeg encoding test on libvpx-vp9."""
    encode("libvpx-vp9", _encode_libvpx_vp9)


def test_encode_libx264() -> None:
    """Perform the ffmpeg encoding test on libx264."""
    encode("libx264", _encode_libx264)


def test_encode_libx265() -> None:
    """Perform the ffmpeg encoding test on libx265."""
    encode("libx265", _encode_libx265)


def test_encode_vp9_vaapi() -> None:
    """Perform the ffmpeg encoding test on vp9_vaapi."""
    encode("vp9_vaapi", _encode_vp9_vaapi)


def test_encode_vvc() -> None:
    """Perform the ffmpeg encoding test on vvc."""
    encode("vvc", _encode_vvc)



"""Perform encoding measures."""

import contextlib
import datetime
import math
import pathlib
import re
import shlex
import shutil
import sqlite3
import subprocess
import tempfile
import uuid

import cutcutcodec
import numpy as np
import orjson
import tqdm
from context_verbose import Printer
from flufl.lock import Lock

from mendevi.cmd import CmdFFMPEG
from mendevi.convert import get_convert_cmd
from mendevi.database.serialize import list_to_binary, tensor_to_binary
from mendevi.ffmpeg_cmd import (
    _encode_av1_nvenc,
    _encode_av1_vaapi,
    _encode_h264_nvenc,
    _encode_h264_vaapi,
    _encode_hevc_nvenc,
    _encode_hevc_vaapi,
    _encode_libaomav1,
    _encode_libopenh264,
    _encode_librav1e,
    _encode_libsvtav1,
    _encode_libvpx_vp9,
    _encode_libx264,
    _encode_libx265,
    _encode_vp9_vaapi,
    _encode_vvc,
)
from mendevi.measures import Activity
from mendevi.utils import best_profile, compute_video_hash, hash_to_signature

ENCODERS_CMD = {
    "av1_nvenc": _encode_av1_nvenc,
    "av1_vaapi": _encode_av1_vaapi,
    "h264_nvenc": _encode_h264_nvenc,
    "h264_vaapi": _encode_h264_vaapi,
    "hevc_nvenc": _encode_hevc_nvenc,
    "hevc_vaapi": _encode_hevc_vaapi,
    "libaom-av1": _encode_libaomav1,
    "libopenh264": _encode_libopenh264,
    "librav1e": _encode_librav1e,
    "libsvtav1": _encode_libsvtav1,
    "libvpx-vp9": _encode_libvpx_vp9,
    "libx264": _encode_libx264,
    "libx265": _encode_libx265,
    "vp9_vaapi": _encode_vp9_vaapi,
    "vvc": _encode_vvc,
}


def encode(src: pathlib.Path, **kwargs: dict) -> tuple[pathlib.Path, str, dict[str]]:
    """Transcode an existing video.

    Parameters
    ----------
    src : pathlib.Path
        The source video file to be transcoded.
    **kwargs : dict
        Transmitted to :py:func:`get_transcode_cmd`.

    Returns
    -------
    dst : pathlib.Path
        The transcoded video path. The stem contains the md5 hash of the file content.
    cmd : str
        The ffmpeg command.
    activity : dict[str]
        The computeur activity during the transcoding process.

    """
    assert isinstance(src, pathlib.Path), src.__class__.__name__
    assert src.is_file(), src

    # find tempfile name
    dst = pathlib.Path(tempfile.gettempdir()) / f"{uuid.uuid4().hex}.mp4"

    # get cmd
    cmd = get_transcode_cmd(src, dst, **kwargs)

    # display
    prt_cmd = " ".join(
        map(shlex.quote, [{str(src): "src.mp4", str(dst): "dst.mp4"}.get(c, c) for c in cmd]),
    )
    with Printer(prt_cmd, color="green") as prt:
        prt.print(f"input video: {src.name}")
        load = tqdm.tqdm(
            dynamic_ncols=True,
            leave=False,
            smoothing=1e-6,
            total=round(float(cutcutcodec.get_duration_video(src)), 2),
            unit="s",
        )

        # transcode
        with Activity() as activity, subprocess.Popen(cmd, stderr=subprocess.PIPE) as process:
            ffmpeg_output = b""
            is_finish = False
            while not is_finish:
                while (
                    match := re.search(
                        br"time=(?P<h>\d+):(?P<m>\d{1,2}):(?P<s>\d{1,2}\.\d*)", ffmpeg_output,
                    )
                ) is None:
                    if not (buff := process.stderr.read(32)):
                        is_finish = True
                        break
                    ffmpeg_output += buff
                else:
                    ffmpeg_output = ffmpeg_output[match.endpos:]
                    elapsed = round(
                        3600.0*float(match["h"]) + 60.0*float(match["m"]) + float(match["s"]),
                        2,
                    )
                    load.total = max(load.total, elapsed)
                    load.update(elapsed-load.n)
            load.close()
            if process.returncode or not dst.stat().st_size:
                dst.unlink(missing_ok=True)
                msg = f"failed to execute {' '.join(map(shlex.quote, cmd))!r}"
                raise RuntimeError(msg)

        # print
        prt.print(f"avg cpu usage: {activity['ps_core']:.1f} %")
        prt.print(f"avg ram usage: {1e-9*np.mean(activity['ps_ram']):.2g} Go")
        if "rapl_power" in activity:
            prt.print(f"avg rapl power: {activity['rapl_power']:.2g} W")
        if "wattmeter_power" in activity:
            prt.print(f"avg wattmeter power: {activity['wattmeter_power']:.2g} W")

        # compute file hash
        signature = hash_to_signature(compute_video_hash(dst, fast=False))
        prt.print(f"output video: sample_{signature}.mp4")

    # move file
    final_dst = src.parent / f"sample_{signature}.mp4"
    if not final_dst.exists():
        shutil.copy(dst, src.parent / f"sample_{signature}_partial.mp4")
        shutil.move(src.parent / f"sample_{signature}_partial.mp4", final_dst)
        final_dst.chmod(0o777)
    dst.unlink()

    return final_dst, prt_cmd, activity


def encode_and_store(
    database: pathlib.Path,
    env_id: int,
    src: pathlib.Path,
    **kwargs: dict,
) -> None:
    """Transcode a video file and store the result in the database.

    Parameters
    ----------
    database : pathlike
        The path of the existing database to be updated.
    env_id : int
        The primary integer key of the environment.
    src : pathlib.Path
        The path of the video to be decoded.
    **kwargs
        Transmitted to :py:func:`encode`.

    Examples
    --------
    >>> import pathlib, tempfile
    >>> from mendevi.database.complete import add_environment
    >>> from mendevi.database.create import create_database
    >>> from mendevi.encode import encode_and_store
    >>> src = pathlib.Path("/data/dataset/video/despacito.mp4")
    >>> create_database(database := pathlib.Path(tempfile.mktemp(suffix=".sqlite")))
    >>> env_id = add_environment(database)
    >>> encode_and_store(
    ...     database, env_id, src,
    ...     encoder="libx264", profile="sd", effort="fast", quality=0.5, threads=8
    ... )
    >>> database.unlink()
    >>>

    """
    # transcode the video
    dst, cmd, activity = encode(src, **kwargs)

    with (
        Lock(str(database.with_name(".dblock")), lifetime=datetime.timedelta(seconds=600)),
        sqlite3.connect(database) as conn,
    ):
        cursor = conn.cursor()

        # fill video table
        with contextlib.suppress(sqlite3.IntegrityError):
            cursor.execute(
                "INSERT INTO t_vid_video (vid_id, vid_name) VALUES (?, ?)",
                (kwargs["src_vid_id"], src.name),
            )
        dst_vid_id: bytes = compute_video_hash(dst)
        with contextlib.suppress(sqlite3.IntegrityError):
            cursor.execute(
                "INSERT INTO t_vid_video (vid_id, vid_name) VALUES (?, ?)",
                (dst_vid_id, dst.name),
            )

        # fill activity table
        activity = {
            "act_duration": activity["duration"],
            "act_gpu_dt": list_to_binary(activity.get("gpu_dt", None)),
            "act_gpu_power": tensor_to_binary(activity.get("gpu_powers", None)),
            "act_ps_core": tensor_to_binary(activity["ps_cores"]),
            "act_ps_dt": list_to_binary(activity["ps_dt"]),
            "act_ps_temp": orjson.dumps(
                activity["ps_temp"], option=orjson.OPT_INDENT_2|orjson.OPT_SORT_KEYS,
            ),
            "act_ps_ram": list_to_binary(activity["ps_ram"]),
            "act_rapl_dt": list_to_binary(activity.get("rapl_dt", None)),
            "act_rapl_power": list_to_binary(activity.get("rapl_powers", None)),
            "act_start": activity["start"],
            "act_wattmeter_dt": list_to_binary(activity.get("wattmeter_dt", None)),
            "act_wattmeter_power": list_to_binary(activity.get("wattmeter_powers", None)),
        }
        keys = list(activity)
        (act_id,) = cursor.execute(
            (
                f"INSERT INTO t_act_activity ({', '.join(keys)}) "
                f"VALUES ({', '.join('?'*len(keys))}) RETURNING act_id"
            ),
            [activity[k] for k in keys],
        ).fetchone()

        # fill encode table
        values = {
            "enc_act_id": act_id,
            "enc_cmd": cmd,
            "enc_dst_vid_id": dst_vid_id,
            "enc_effort": kwargs["effort"],
            "enc_encoder": kwargs["encoder"],
            "enc_env_id": env_id,
            "enc_fps": float(kwargs["fps"]),
            "enc_height": kwargs["resolution"][0],
            "enc_pix_fmt": kwargs["pix_fmt"],
            "enc_quality": kwargs["quality"],
            "enc_src_vid_id": kwargs["src_vid_id"],
            "enc_threads": kwargs["threads"],
            "enc_mode": kwargs["mode"],
            "enc_width": kwargs["resolution"][1],
        }
        keys = list(values)
        cursor.execute(
            f"INSERT INTO t_enc_encode ({', '.join(keys)}) VALUES ({', '.join('?'*len(keys))})",
            [values[k] for k in keys],
        )


def get_transcode_cmd(src: pathlib.Path, dst: pathlib.Path, **kwargs: dict) -> list[str]:
    """Return the ffmpeg encode cmd."""
    # header
    cmd: list[str] = ["ffmpeg", "-y", "-i", str(src)]

    # filter
    if (
        filter_cmd := get_convert_cmd(
            src,
            additional_filter=kwargs["filter"],
            fps=kwargs["fps"],
            pix_fmt=kwargs["pix_fmt"],
            resolution=kwargs["resolution"],
        )
    ):
        cmd.extend(["-vf", filter_cmd])

    # transcode
    cmd.extend(ENCODERS_CMD[kwargs["encoder"]](**kwargs))
    # mount in RAM: mount -o mode=1777,nosuid,nodev -t tmpfs tmpfs /tmp
    # # https://trac.ffmpeg.org/wiki/Encode/AV1#AMDAMFAV1
    # # https://trac.ffmpeg.org/wiki/Hardware/AMF
    # # https://github.com/GPUOpen-LibrariesAndSDKs/AMF/wiki/GPU%20and%20APU%20HW%20Features%20and%20Support
    # case "h264_amf":  # for AMD GPU
    #     raise NotImplementedError
    # case "av1_amf":  # for AMD GPU
    #     raise NotImplementedError

    # final
    cmd.append(str(dst))
    return cmd


def quality_to_rate(kwargs: dict[str]) -> int:
    """Return the absolute target bitrate in kbit/s.

    Based on https://twitch-overlay.fr/quelle-connexion-internet-choisir-pour-streamer-sur-twitch/
    and https://bitmovin.com/blog/video-bitrate-streaming-hls-dash/

    You can plot the bitrate with: mendevi plot mendevi.db -x bitrate -y psnr -f 'mode = "vbr"'

    The flow margin is taken to be twice as small and twice as large as the recommendations.
    """
    quality = kwargs["quality"]
    assert isinstance(quality, float), quality.__class__.__name__
    assert 0.0 <= quality <= 1.0, quality
    match (profile := best_profile(*kwargs["resolution"])):
        case "sd":
            mini, maxi = 400, 2100
        case "hd":
            mini, maxi = 1500, 6000
        case "fhd":
            mini, maxi = 3000, 9000
        case "uhd4k":
            mini, maxi = 10000, 51000
        case _:
            msg = f"please define a bitrate rule for the profile {profile}"
            raise NotImplementedError(msg)
    mini, maxi = mini // 2, maxi * 2  # apply margin
    mini, maxi = math.log10(float(mini)), math.log10(float(maxi))
    return round(10.0**(maxi-quality*(maxi-mini)))



"""Fill the SQL database."""

import datetime
import pathlib
import sqlite3
import time

import context_verbose as cv
import numpy as np
import orjson
from flufl.lock import Lock

from mendevi.measures import Activity
from mendevi.measures.context import full_context

from .create import ENV_UNIQUE, is_sqlite
from .serialize import list_to_binary, tensor_to_binary


def get_idle(duration: float = 60.0) -> dict:
    """Measure the idle consumption.

    Parameters
    ----------
    duration : float, default=60
        The total duration of the measurement, bloking time in second.

    Returns
    -------
    measures : dict
        Same as :py:class:`mendevi.measures.Activity`.

    Examples
    --------
    >>> from mendevi.database.complete import get_idle
    >>> idle = get_idle(duration=1.0)  # doctest: +ELLIPSIS
    Measure IDLE during 1.0 seconds...
    █ avg cpu usage: ... %
    █ avg ram usage: ... Go
    █ avg rapl power: ... W
    >>>

    """
    assert isinstance(duration, float), duration.__class__.__name__
    assert duration > 0.0, duration.__class__.__name__

    with cv.Printer(f"Measure IDLE during {duration} seconds...") as printer:

        # measure IDLE
        with Activity() as activity:
            time.sleep(duration)

        # print
        printer.print(f"avg cpu usage: {activity['ps_core']:.1f} %")
        printer.print(f"avg ram usage: {1e-9*np.mean(activity['ps_ram']):.2g} Go")
        if "rapl_power" in activity:
            printer.print(f"avg rapl power: {activity['rapl_power']:.2g} W")
        if "gpu_power" in activity:
            printer.print(f"avg GPU power: {activity['gpu_power']:.2g} W")
        if "wattmeter_power" in activity:
            printer.print(f"avg wattmeter power: {activity['wattmeter_power']:.2g} W")

    return activity


def add_environment(database: sqlite3.Cursor | str | bytes | pathlib.Path) -> int:
    """Complete the environment table with the current environment.

    Parameters
    ----------
    database : pathlike
        The path of the existing database to be updated.

    Returns
    -------
    id : int
        The primary key of the freshly updated environment table.

    """
    # open file
    if not isinstance(database, sqlite3.Cursor):
        database = pathlib.Path(database).expanduser()
        assert is_sqlite(database), database
        with (
            Lock(str(database.with_name(".dblock")), lifetime=datetime.timedelta(seconds=1800)),
            sqlite3.connect(database) as conn,
        ):
            cursor = conn.cursor()
            key = add_environment(cursor)
            cursor.close()
            return key

    # try to set the context
    environment = full_context()
    environment = {f"env_{k}": v for k, v in environment.items()}
    keys = list(environment)  # to keep order
    try:
        (env_id,) = database.execute(
            (
                f"INSERT INTO t_env_environment ({', '.join(keys)}) "
                f"VALUES ({', '.join('?'*len(keys))}) RETURNING env_id"
            ),
            [environment[k] for k in keys],
        ).fetchone()
    except sqlite3.IntegrityError:
        (env_id,) = database.execute(
            (
                "SELECT env_id FROM t_env_environment "
                f"WHERE {' AND '.join(f'{k}=?' for k in ENV_UNIQUE)}"
            ),
            [environment[k] for k in ENV_UNIQUE],
        ).fetchone()
        return env_id

    # get idle
    idle = get_idle()
    idle = {
        "act_duration": idle["duration"],
        "act_ps_core": tensor_to_binary(idle["ps_cores"]),
        "act_ps_dt": list_to_binary(idle["ps_dt"]),
        "act_ps_temp": orjson.dumps(
            idle["ps_temp"], option=orjson.OPT_INDENT_2|orjson.OPT_SORT_KEYS,
        ),
        "act_ps_ram": list_to_binary(idle["ps_ram"]),
        "act_rapl_dt": list_to_binary(idle.get("rapl_dt", None)),
        "act_rapl_power": list_to_binary(idle.get("rapl_powers", None)),
        "act_start": idle["start"],
        "act_wattmeter_dt": list_to_binary(idle.get("wattmeter_dt", None)),
        "act_wattmeter_power": list_to_binary(idle.get("wattmeter_powers", None)),
    }
    keys = list(idle)
    (act_id,) = database.execute(
        (
            f"INSERT INTO t_act_activity ({', '.join(keys)}) "
            f"VALUES ({', '.join('?'*len(keys))}) RETURNING act_id"
        ),
        [idle[k] for k in keys],
    ).fetchone()
    database.execute(
        (
            f"UPDATE t_env_environment SET env_idle_act_id={act_id} "
            f"WHERE {' AND '.join(f'{k}=?' for k in ENV_UNIQUE)}"
        ),
        [environment[k] for k in ENV_UNIQUE],
    )
    return env_id



"""Benchmark."""

import pathlib
import sqlite3
import tempfile
import time

import pytest

from mendevi.database.create import create_database


def test_unicity_activity_from_env() -> None:
    """Ensure it raise and exception."""
    create_database(database := tempfile.mktemp(suffix=".sqlite"))

    # create 4 activities and a video
    with sqlite3.connect(database) as conn:
        for _ in range(4):
            conn.execute(
                "INSERT INTO t_act_activity (act_start, act_duration) VALUES (?, ?)",
                (time.time(), 1.0),
            )
        conn.execute(
            "INSERT INTO t_vid_video (vid_id) VALUES (?)",
            ((0).to_bytes(16),),
        )

    # insert encode and decode
    with sqlite3.connect(database) as conn:
        conn.execute(
            "INSERT INTO t_enc_encode (enc_dst_vid_id, enc_env_id) VALUES (?, ?)",
            ((0).to_bytes(16), 0),
        )
        conn.execute(
            "INSERT INTO t_dec_decode (dec_vid_id, dec_env_id) VALUES (?, ?)",
            ((0).to_bytes(16), 1),
        )

    # insert environment OK
    with sqlite3.connect(database) as conn:
        conn.execute(
            """
            INSERT INTO t_env_environment
            (env_ffmpeg_version, env_hostname, env_logical_cores, env_ram, env_idle_act_id)
            VALUES (?, ?, ?, ?, ?)
            """,
            ("ffmpeg 0.0.0", "my_name", 8, 1024, 2),
        )

    # insert environnement FAIL
    with sqlite3.connect(database) as conn, pytest.raises(sqlite3.IntegrityError):
        conn.execute(
            """
            INSERT INTO t_env_environment
            (env_ffmpeg_version, env_hostname, env_logical_cores, env_ram, env_idle_act_id)
            VALUES (?, ?, ?, ?, ?)
            """,
            ("ffmpeg 0.0.1", "my_name", 8, 1024, 2),
        )

    # insert environment OK
    with sqlite3.connect(database) as conn:
        conn.execute(
            """
            INSERT INTO t_env_environment
            (env_ffmpeg_version, env_hostname, env_logical_cores, env_ram, env_idle_act_id)
            VALUES (?, ?, ?, ?, ?)
            """,
            ("ffmpeg 0.0.2", "my_name", 8, 1024, 3),
        )

    # insert environment OK null env_idle_act_id
    with sqlite3.connect(database) as conn:
        conn.execute(
            """
            INSERT INTO t_env_environment
            (env_ffmpeg_version, env_hostname, env_logical_cores, env_ram)
            VALUES (?, ?, ?, ?)
            """,
            ("ffmpeg 0.0.3", "my_name", 8, 1024),
        )

    # insert environment OK null env_idle_act_id second time
    with sqlite3.connect(database) as conn:
        conn.execute(
            """
            INSERT INTO t_env_environment
            (env_ffmpeg_version, env_hostname, env_logical_cores, env_ram)
            VALUES (?, ?, ?, ?)
            """,
            ("ffmpeg 0.0.4", "my_name", 8, 1024),
        )

    # update environment FAIL
    with sqlite3.connect(database) as conn, pytest.raises(sqlite3.IntegrityError):
        conn.execute(
            """
            UPDATE t_env_environment SET env_idle_act_id=3
            WHERE env_ffmpeg_version=? AND env_hostname=? AND env_logical_cores=? AND env_ram=?
            """,
            ("ffmpeg 0.0.3", "my_name", 8, 1024),
        )

    pathlib.Path(database).unlink()



"""Test special issues of merge_extractors."""

import pytest

from mendevi.database.meta import merge_extractors


def test_alias_cycle() -> None:
    """Ensure it raises ValueError if the alias are cycling."""
    with pytest.raises(ValueError, match=r".*graph has cycles.*"):
        merge_extractors({"foo"}, alias={"foo": "toto", "toto": "foo"})


def test_not_defined() -> None:
    """Ensure it raises ValueError if a variable is not defined."""
    with pytest.raises(ValueError, match=r".*graph has cycles.*"):
        merge_extractors({"foo"})



"""All measurement helper."""

import numbers
import platform
import queue
import threading
import time

from .g5kpower import g5kpower
from .gpu import UsageGPU
from .psutil import Usage
from .rapl import RAPL

G5K_API_WAIT = 2.0  # waiting time (in seconds) before data from the grid5000 API becomes accessible


class Activity(threading.Thread):
    """Measure the computer activity of a section.

    Examples
    --------
    >>> import pprint, time
    >>> from mendevi.measures import Activity
    >>> with Activity() as activity:
    ...     time.sleep(1)
    ...
    >>>

    """

    def __init__(self, sleep: numbers.Real = 50e-3) -> None:
        """Init the perf context.

        Parameters
        ----------
        sleep : float, default=50e-3
            The time interval between 2 measures (in s).

        """
        assert isinstance(sleep, numbers.Real), sleep.__class__.__name__
        assert sleep > 0, sleep

        super().__init__(daemon=True)

        self._rapl_catcher = RAPL(sleep=sleep, no_fail=True)
        self._usage_catcher = Usage(sleep=sleep)
        self._usage_gpu_catcher = UsageGPU(sleep=sleep)
        self._exit_queue = queue.Queue()
        self.sleep = float(sleep)
        self.res: dict = {}

    def run(self) -> None:
        """Perform the measures."""
        self.res["start"] = time.time()
        with (
            self._rapl_catcher as rapl,
            self._usage_catcher as usage,
            self._usage_gpu_catcher as gpu,
        ):
            self._exit_queue.get()  # wait
        self.res |= {
            "ps_core": usage["cpu"],
            "ps_cores": usage["cpus"],
            "ps_dt": usage["dt"],
            "ps_ram": usage["ram"],
            "ps_temp": usage["temp"],
        }
        if rapl is not None:
            self.res |= {
                "rapl_dt": rapl["dt"],
                "rapl_energy": rapl["energy"],
                "rapl_power": rapl["power"],
                "rapl_powers": rapl["powers"],
            }
        if gpu is not None:
            self.res |= {
                "gpu_cores": gpu["gpus"],
                "gpu_dt": gpu["dt"],
                "gpu_energy": gpu["energy"],
                "gpu_memory": gpu["memory"],
                "gpu_power": gpu["power"],
                "gpu_powers": gpu["powers"],
            }

    def __enter__(self) -> dict:
        r"""Start to measure.

        Returns
        -------
        activity: dict[str]
            * duration: float, the real measure duration.
            * gpu_cores: list[list[float]], tensor of detailed usage of each GPU in [0, 1].
            * gpu_dt: list[float], the duration of each interval (in s).
            * gpu_energy: float, the total energy consumption of all GPU (in J).
            * gpu_power: float, the total power consumption of all GPU (in W).
            * gpu_powers: list[list[float]], the sampled power in each point (in W).
            * ps_core: float, the mean cummulated usage of all the logical cpus.
            * ps_cores: list[list[float]], tensor of detailed usage of each logical core in %.
            * ps_dt: list[float], the duration of each interval (in s).
            * ps_ram: list[int], list of the sampled ram usage in bytes in each point.
            * ps_temp: dict[str, lis[float]], the temperature of each device (in C).
            * rapl_dt: list[float], the duration of each interval (in s).
            * rapl_energy: float, the total energy consumption (in J).
            * rapl_power: float, the average power, energy divided by the duration (in W).
            * rapl_powers: list[float], the average power in each interval (in W).
            * start: float, absolute timestamp.
            * wattmeter_dt: list[float], the duration of each interval (in s).
            * wattmeter_energy: float, the total energy consumption (in J).
            * wattmeter_power: float, the average power, energy divided by the duration (in W).
            * wattmeter_powers: list[float], the sampled power in watt in each point.

        Notes
        -----
        The returned dictionary is update inplace when we exit the code bloc.
        Only the successfull field are created.

        """
        self.start()
        return self.res

    def __exit__(self, *_: object) -> None:
        """Stop the measure and update the dictionary returnd by __enter__."""
        # stop
        self.res["duration"] = time.time() - self.res["start"]
        self._exit_queue.put(None)
        self.join()
        # request wattmeter power
        time.sleep(max(0.0, G5K_API_WAIT + self.res["start"] + self.res["duration"] - time.time()))
        try:
            wattmeter = g5kpower(platform.node(), self.res["start"], self.res["duration"])
        except ValueError:
            wattmeter = None
        else:
            self.res |= {
                "wattmeter_dt": wattmeter["dt"],
                "wattmeter_energy": wattmeter["energy"],
                "wattmeter_power": wattmeter["power"],
                "wattmeter_powers": wattmeter["powers"],
            }



"""Try to read the energy with RAPL."""

import logging
import numbers
import os
import re
import subprocess
import sys
import time

PATTERN = re.compile(br"^\s*(?P<time>\d+\.\d*);(?P<energy>\d+[.,]\d*);(?P<unit>Joules)")


class RAPL:
    """Uses the linux perf command through a python context manager.

    Examples
    --------
    >>> import time
    >>> from mendevi.measures.rapl import RAPL
    >>> with RAPL() as energy:
    ...     time.sleep(1)
    ...
    >>>

    """

    def __init__(self, sleep: numbers.Real = 50e-3, *, no_fail: bool = False) -> None:
        """Init the perf context.

        Parameters
        ----------
        sleep : float, default=50e-3
            The time interval between 2 measures (in s).
        no_fail : bool, default=True
            If False, raise RuntimeError if it fails to get the RAPL measure.
            Otherwise (if True), return None instead of failing.

        """
        assert isinstance(sleep, numbers.Real), sleep.__class__.__name__
        assert sleep > 0, sleep
        assert isinstance(no_fail, bool), no_fail

        self.sleep = round(1000*float(sleep))  # sleep time in ms
        self.res: dict = {"dt": [], "energy": None, "power": None, "powers": []}
        self.process = subprocess.Popen(  # pylint: disable=R1732
            [  # sudo apt install linux-perf
                "perf", "stat",
                "--event", "power/energy-pkg/",  # cores and cache, no ram
                "--all-cpus",
                "--field-separator", ";",  # output csv like
                "--interval-print", str(self.sleep),
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        self.lines = [self.process.stderr.readline()]
        self.time_bounds = [time.time(), None, None]  # start proc, measure, stop
        if PATTERN.search(self.lines[0]) is None:
            if os.geteuid() != 0:  # if not root
                logging.getLogger(__name__).error("try with 'sudo %s'", sys.executable)
                logging.getLogger(__name__).error(
                    "or give rights with 'sudo sysctl -w kernel.perf_event_paranoid=0'",
                )
                logging.getLogger(__name__).error(
                    "or edit /etc/sysctl.d/99-perf.conf with 'kernel.perf_event_paranoid = 0'",
                )
            self.lines.extend(self.process.stderr.readlines())
            if not no_fail:
                raise RuntimeError(b"".join(self.lines).decode())
            logging.getLogger(__name__).error(b"".join(self.lines).decode())
            self.res = None

    def __enter__(self) -> dict:
        """Start to measure.

        Returns
        -------
        Consumption: dict[str]
            * 'dt': The time difference between 2 consecutive power measurements (in s).
            * 'energy': The total energy consumption (in J).
            * 'power': The average power, energy divided by the duration (in w).
            * 'powers': The power measured between 2 consecutive points (in w).

        """
        self.time_bounds[1] = time.time()
        return self.res

    def __exit__(self, *_: object) -> None:
        """Stop the measure and update the dictionary returnd by __enter__."""
        # stop measuring
        self.time_bounds[2] = time.time()
        time.sleep(self.sleep/1000)  # to be shure catching the last point
        self.process.terminate()
        self.lines.extend(self.process.stderr.readlines())

        # exit if failed
        if self.res is None:
            return

        # decode output
        values = [PATTERN.search(line) for line in self.lines]
        times_end, energy = zip(
            *(
                (float(p["time"]), float(p["energy"].replace(b",", b".")))
                for p in values if p is not None
            ), strict=False,
        )

        # convert energy in power
        times_end = [t - times_end[0] + self.time_bounds[0] for t in times_end]
        times_start = [times_end[0] - self.sleep / 1000, *times_end[:-1]]
        powers = [e / (te - ts) for e, ts, te in zip(energy, times_start, times_end, strict=False)]

        # pad and crop
        while times_end[-1] < self.time_bounds[2]:  # pad
            times_end.append(times_end[-1] + self.sleep/1000)
            times_start.append(times_end[-2])
            powers.append(powers[-1])  # assumption: cst interpolation
        while times_end[0] < self.time_bounds[1]:  # crop start
            del times_start[0], times_end[0], powers[0]
        while times_start[-1] > self.time_bounds[2]:  # crop end
            del times_start[-1], times_end[-1], powers[-1]
        times_start[0], times_end[-1] = self.time_bounds[1], self.time_bounds[2]

        # compute total energy
        self.res["dt"] = [te - ts for ts, te in zip(times_start, times_end, strict=False)]
        self.res["energy"] = sum(p * dt for p, dt in zip(powers, self.res["dt"], strict=False))
        self.res["power"] = self.res["energy"] / (times_end[-1] - times_start[0])
        self.res["powers"] = powers



"""Unify all the models with a common sctucture."""

import abc
import logging
import pathlib
import sqlite3
import typing

import numpy as np
import torch
from context_verbose import Printer

from mendevi.cst.labels import LABELS
from mendevi.database.create import is_sqlite
from mendevi.database.extract import SqlLinker
from mendevi.database.meta import get_extractor, merge_extractors
from mendevi.download.decapsulation import retrive_file


class Model(abc.ABC):
    """Common structure to all models.

    Attributes
    ----------
    cite : str
        The latex bibtext model citation.
    parameters : torch.Tensor | None
        The trainable parameters of the model (read and write).
    input_labels : list[str]
        The name of all input parameters (readonly).
    output_labels : list[str]
        The name of all output parameters (readonly).

    """

    def __init__(self, title: str | None = None, **kwargs: dict) -> None:
        """Initialise the model.

        Parameters
        ----------
        title : str, optional
            The model title.
        **kwargs : dict
            Includes the following fields.
        sources : str
            All sources for the model, the conference paper, the authors, etc.
        input_labels : list[str]
            The name of all input parameters.
            The possibles values are :py:cst:`mendevi.plot.axis.Name`.
        output_labels : list[str]
            The name of all output parameters.
            The possibles values are :py:cst:`mendevi.plot.axis.Name`.
        parameters : object, optional
            The learnable parameters for regressive models.

        """
        assert set(kwargs).issubset({"sources", "input_labels", "output_labels", "parameters"})
        # check input_labels
        input_labels = kwargs.get("input_labels", [])
        assert hasattr(input_labels, "__iter__"), input_labels.__class__.__name__
        input_labels = list(input_labels)
        assert input_labels, "input must be not empty"
        assert all(isinstance(lab, str) and lab in LABELS for lab in input_labels), input_labels
        self._input_labels = input_labels

        # check output_labels
        output_labels = kwargs.get("output_labels", [])
        assert hasattr(output_labels, "__iter__"), output_labels.__class__.__name__
        output_labels = list(output_labels)
        assert output_labels, "output must be not empty"
        assert all(isinstance(lab, str) and lab in LABELS for lab in output_labels), output_labels
        self._output_labels = output_labels

        # check parameters
        self._parameters = kwargs.get("parameters")

        # check title
        if title is None:
            title = (
                f"{'regressive ' if self._parameters is not None else ''}model "
                f"to predict {', '.join(sorted(self._output))} "
                f"from {', '.join(sorted(self._input))}"
            )
        else:
            assert isinstance(title, str), title.__class__.__name__
        self._title = title

        # check authors
        sources = kwargs.get("sources", "")
        assert isinstance(sources, str), sources.__class__.__name__
        self._sources = sources

    def _fit(self, values: dict[str]) -> typing.Never:
        """Perform regression on parameters ``self.parameters``."""
        raise NotImplementedError

    @abc.abstractmethod
    def _predict(self, values: dict[str]) -> dict[str]:
        """Implement the heart of the model."""
        raise NotImplementedError

    @property
    def cite(self) -> str:
        """Return the bibtex citation."""
        raise NotImplementedError

    def fit(
        self,
        database: pathlib.Path | str,
        select: str | None = None,
        query: str | None = None,
    ) -> typing.Self:
        """Fit the trainable hyper parameters of the models.

        Parameters
        ----------
        database : pathlike
            The training database.
        select : str, optional
            The python expression to keep the line, like ``mendevi plot --filter``.
        query : str, optional
            If provided, use this sql query to perform the request,
            otherwise (default) attemps to guess the query.

        Return
        ------
        self
            A reference to the inplace fitted model.

        """
        with Printer(f"Fit {self._title!r}...", color="pink") as prt:
            # verification
            database = retrive_file(database)
            assert is_sqlite(database), f"{database} is not a valid SQL database"

            # get sql query
            prt.print("get SQL query")
            atom_names, line_extractor = merge_extractors(
                set(self._input_labels) | set(self._output_labels),
                select=select,
                return_callable=True,
            )
            if query is None:
                select = {s for lbl in atom_names for s in get_extractor(lbl).func.select}
                if len(queries := SqlLinker(*select).sql) == 0:
                    msg = "fail to create the SQL query, please provide it yourself"
                    raise RuntimeError(msg)
                if len(queries) > 1:
                    logging.getLogger(__name__).warning(
                        "several request founded %s, please select it yourself", queries,
                    )
                query = queries.pop(0)
            else:
                assert isinstance(query, str), query.__class__.__name__

            # perform sql request
            prt.print("perform SQL query")
            values = {label: [] for label in set(self._input_labels) | set(self._output_labels)}
            with sqlite3.connect(database) as conn:
                conn.row_factory = sqlite3.Row
                for raw in conn.execute(query):
                    for label, value in line_extractor(dict(raw)).items():
                        values[label].append(value)

            # fit the model
            prt.print("fit the model")
            self._fit(values)

            prt.print_time()
        return self

    @property
    def input_labels(self) -> list[str]:
        """Return the name of all input parameters."""
        return self._input_labels.copy()

    @property
    def output_labels(self) -> list[str]:
        """Return the name of all output parameters."""
        return self._output_labels.copy()

    @property
    def parameters(self) -> torch.Tensor or None:
        """Return the trainable parameters of the model."""
        return self._parameters

    @parameters.setter
    def parameters(self, new_params: torch.Tensor) -> None:
        """Update the parameters."""
        if self._parameters is not None and new_params.__class__ != self._parameters.__class__:
            logging.getLogger(__name__).warning("chage the type of parameters")
        self._parameters = new_params

    def predict(self, *input_args: tuple, **input_kwargs: dict) -> dict[str]:
        """Perform the prediction(s) of this model.

        Parameters
        ----------
        *input_args, **input_kwargs
            The parameters values, with the keys defined during initialisation.

        Returns
        -------
        prediction : dict[str]
            Associate each ouput variable with the prediction.

        """
        with Printer(f"Predict {self._title!r}...", color="pink") as prt:
            # check args
            prt.print("check args")
            values: dict[str] = {}
            for i, arg in enumerate(input_args):
                assert i != len(self._input_labels), (
                    f"only {len(self._input)} arguments expeted {self._input_labels}, "
                    f"{input_args} given"
                )
                values[self._input_labels[i]] = arg
            for name, arg in input_kwargs.items():
                if name in values:
                    msg = f"argument {name} given twice"
                    raise ValueError(msg)
                if name not in self._input:
                    msg = f"only {self._input_labels} arguments excpected, not {name}"
                    raise ValueError(msg)
                values[name] = arg

            # cast args
            prt.print("cast args")
            for name, arg in values.copy().items():
                match arg:
                    case float():
                        values[name] = torch.asarray(arg, dtype=torch.float32)
                    case np.ndarray():
                        values[name] = torch.from_numpy(arg)
                    case list():
                        if all(isinstance(item, float) for item in arg):
                            values[name] = torch.asarray(arg, dtype=torch.float32)
                prt.print(f"{name} = {values[name]!s:.80}")

            # predict
            prt.print("predict")
            prediction = self._predict(values)

            # check output
            assert isinstance(prediction, dict), prediction.__class__.__name__
            assert prediction.keys() == set(self._output_labels), \
                f"_predict must return {self._output_labels}, not {sorted(prediction)}"

            prt.print_time()
        return prediction



"""Serialize the sql binary data."""

import functools
import struct

import numpy as np

VECT_DTYPE = np.float32


def skip_none(func: callable) -> callable:
    """Allow skipping the process of None input."""
    @functools.wraps(func)
    def decorated_func(arg: object | None) -> object | None:
        return None if arg is None else func(arg)
    return decorated_func


@skip_none
def list_to_binary(vect: list[float]) -> bytes:
    r"""Serialize a float vector into binary data.

    Bijection of :py:func:`binary_to_list`.

    Parameters
    ----------
    vect : arralike
        The list of floats

    Returns
    -------
    data : bytes
        The serialized float32 list.

    Examples
    --------
    >>> from mendevi.database.serialize import list_to_binary
    >>> list_to_binary([0.0, -1.0, 0.123456789])
    b'\x00\x00\x00\x00\x00\x00\x80\xbf\xea\xd6\xfc='
    >>>

    """
    vect = np.asarray(vect, dtype=VECT_DTYPE)
    assert vect.ndim == 1, vect.shape
    return vect.tobytes("C")


@skip_none
def binary_to_list(data: bytes) -> np.ndarray[VECT_DTYPE]:
    r"""Deserialize a binary data into a float vector.

    Bijection of :py:func:`list_to_binary`.

    Parameters
    ----------
    data : bytes
        The serialized float 32 list.

    Returns
    -------
    vect : np.ndarray
        The list of floats

    Examples
    --------
    >>> from mendevi.database.serialize import binary_to_list
    >>> binary_to_list(b'\x00\x00\x00\x00\x00\x00\x80\xbf\xea\xd6\xfc=')
    array([ 0.        , -1.        ,  0.12345679], dtype=float32)
    >>>

    """
    assert isinstance(data, bytes), data.__class__.__name__
    assert len(data) % np.dtype(VECT_DTYPE).itemsize == 0, len(data)
    return np.frombuffer(data, dtype=VECT_DTYPE, count=len(data)//np.dtype(VECT_DTYPE).itemsize)


@skip_none
def tensor_to_binary(tensor: list[list[float]]) -> bytes:
    r"""Serialize a 2d array into binary data.

    Bijection os :py:func:`binary_to_tensor`.

    Parameters
    ----------
    tensor : arraylike
        The list of list of float

    Returns
    -------
    data : bytes
        The serialized float32 tensor.

    Examples
    --------
    >>> from mendevi.database.serialize import tensor_to_binary
    >>> tensor_to_binary([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6]])
    b'\x02\x00\x00\x00\xcd\xcc\x8c?\xcd\xcc\x0c@33S@\xcd\xcc\x8c@\x00\x00\xb0@33\xd3@'
    >>>

    """
    tensor = np.asarray(tensor, dtype=VECT_DTYPE)
    assert tensor.ndim == 2, tensor.shape
    return struct.pack("I", tensor.shape[1]) + tensor.tobytes("C")


@skip_none
def binary_to_tensor(data: bytes) -> np.ndarray[VECT_DTYPE, VECT_DTYPE]:
    r"""Serialize a 2d array into binary data.

    Bijection os :py:func:`binary_to_tensor`.

    Parameters
    ----------
    data : bytes
        The serialized float32 tensor.

    Returns
    -------
    tensor : np.ndarray
        The list of list of float

    Examples
    --------
    >>> from mendevi.database.serialize import binary_to_tensor
    >>> data = b'\x02\x00\x00\x00\xcd\xcc\x8c?\xcd\xcc\x0c@33S@\xcd\xcc\x8c@\x00\x00\xb0@33\xd3@'
    >>> binary_to_tensor(data)
    array([[1.1, 2.2],
           [3.3, 4.4],
           [5.5, 6.6]], dtype=float32)
    >>>

    """
    assert isinstance(data, bytes), data.__class__.__name__
    size, data = data[:4], data[4:]
    return np.frombuffer(data, dtype=VECT_DTYPE).reshape(-1, struct.unpack("I", size)[0])



"""Perform decoding measures."""

import hashlib
import itertools
import sqlite3
from typing import TYPE_CHECKING

import click
from context_verbose import Printer

from mendevi.database.complete import add_environment
from mendevi.decode import decode_and_store
from mendevi.utils import compute_video_hash, get_resolution

from .parse import ResolutionParamType, parse_videos_database

if TYPE_CHECKING:
    import pathlib


def _parse_args(prt: Printer, kwargs: dict) -> None:
    """Verification of the arguments."""
    # repeat
    assert "repeat" in kwargs, sorted(kwargs)
    assert isinstance(kwargs["repeat"], int), kwargs["repeat"].__class__.__name__
    assert kwargs["repeat"] >= 1, kwargs["repeat"]
    prt.print(f"repeat    : {kwargs['repeat']}")

    # resolution
    assert "resolution" in kwargs, sorted(kwargs)
    assert isinstance(kwargs["resolution"], tuple), kwargs["resolution"].__class__.__name__
    assert all(
        isinstance(r, tuple) and len(r) == 2 and all(isinstance(s, int) and s > 0 for s in r)
        for r in kwargs["resolution"]
    ), kwargs["resolution"]
    if kwargs["resolution"]:
        prt.print(f"resolution: {', '.join(f'w={w}:h={h}' for h, w in kwargs['resolution'])}")
    else:
        kwargs["resolution"] = (None,)

        # filter
    kwargs["filter"] = kwargs.get("filter", ())
    assert isinstance(kwargs["filter"], tuple), kwargs["filter"].__class__.__name__
    assert all(isinstance(f, str) for f in kwargs["filter"]), kwargs["filter"]
    if kwargs["filter"]:
        prt.print(f"filters   : {', '.join(map(str, kwargs['filter']))}")
    else:
        kwargs["filter"] = ("",)


@click.command()
@click.argument("videos", nargs=-1, type=click.Path())
@click.option("-d", "--database", type=click.Path(), help="The database path.")
@click.option(
    "-r", "--repeat",
    type=int,
    default=5,
    help="The number of times the decoding is repeated on this machine.",
)
@click.option(
    "--resolution",
    type=ResolutionParamType(),
    multiple=True,
    help="The optional video shape conversion during decoding.",
)
@click.option(
    "--filter",
    type=str,
    multiple=True,
    help="The ffmpeg filter to apply on fly.",
)
def main(videos: tuple[str], database: str | None = None, **kwargs: dict) -> None:
    """Measures activity during decoding.

    \b
    Parameters
    ----------
    videos : tuple[pathlike], optional
        All videos to be decoded. It can be a glob expression, a directory or a file path.
    database : pathlike, optional
        The path to the database where all measurements are stored.
        If a folder is provided, the database is created inside this folder.
    **kwargs: dict
        Please refer to the detailed arguments below.
    repeat : int, default=5
        The number of times the experiment is repeated on this environment.
        This allows us to estimate the variance of measurements.
    resolution : tuple[int, int], optional
        If provided, the decoded video will be reshaped on fly.
        If this argument is not provided, the resolution remains unchanged.
    filter : str, optional
        The additional ffmpeg video filter (after the -vf command)
        to be applied immediately after decoding and before other conversions.

    """
    with Printer("Parse configuration...") as prt:
        videos, database = parse_videos_database(prt, videos, database)
        _parse_args(prt, kwargs)

    # preparation of context
    env_id = add_environment(database)
    kwargs["dec_vid_id"]: dict[pathlib.Path, bytes] = compute_video_hash(videos)

    # retrieves the settings for videos that have already been decoded
    with sqlite3.connect(database) as conn:
        conn.row_factory = sqlite3.Row
        done: dict[tuple, int] = {}
        for row in conn.execute(
            "SELECT * FROM t_dec_decode WHERE dec_env_id=?", (env_id,),
        ):
            values = (
                row["dec_vid_id"],
                row["dec_filter"] or "",
                (row["dec_height"], row["dec_width"]),
            )
            done[values] = done.get(values, 0) + 1
        prt.print(f"{sum(done.values())} video already decoded under these environment")

    # iterate on all the parameters
    loops = sorted(
        itertools.product(videos, kwargs["filter"], kwargs["resolution"]),
        key=lambda t: hashlib.md5(str(t).encode("utf-8")).hexdigest(),  # repetable shuffle
    )
    for i, (repeat, values_) in enumerate(itertools.product(range(kwargs["repeat"]), loops)):
        values = dict(zip(("video", "filter", "resolution"), values_, strict=True))
        values["repeat"] = repeat
        values["dec_vid_id"] = kwargs["dec_vid_id"][values["video"]]
        values["filter"] = values["filter"] or ""
        values["resolution"] = values["resolution"] or get_resolution(values["video"])
        key = (values["dec_vid_id"], values["filter"], values["resolution"])
        if done.get(key, 0) > values["repeat"]:
            continue
        with Printer(f"Decode {i+1}/{kwargs["repeat"]*len(loops)}...", color="cyan") as prt:
            decode_and_store(database, env_id, values.pop("video"), **values)
            done[key] = done.get(key, 0) + 1
            prt.print_time()



"""Generate a Python code that allows you to plot the graph."""

import datetime
import itertools
import logging
import re
import subprocess

from mendevi import __version__
from mendevi.cst.plot import COLORS, FIGSIZE, MARKERS, MARKERSIZE
from mendevi.database.extract import SqlLinker
from mendevi.database.meta import Scale, get_extractor, merge_extractors


def printer(**kwargs: dict) -> str:
    """Create an excecutable python code."""
    # create code
    code: list[str] = []
    code.extend(print_header())
    code.extend(print_import())
    code.extend(print_cst(**kwargs))
    code.extend(print_mean_std())
    code.extend(print_read_sql(**kwargs))
    code.extend(print_fill_axe(**kwargs))
    code.extend(print_main(**kwargs))
    code.extend(print_entry())
    code = "\n".join(code)

    # format code
    try:
        result = subprocess.run(
            ["ruff", "format", "-"],
            input=code.encode(),
            capture_output=True,
            check=True,
        )
        return result.stdout.decode()
    except FileNotFoundError:
        logging.getLogger(__name__).warning(
            "Ruff is not installed, please install it (uv pip install ruff)",
        )
    except subprocess.CalledProcessError as err:
        logging.getLogger(__name__).warning("Ruff formatting failed: %s", err.stderr.decode())
    return code


def print_entry() -> list[str]:
    """Return the code of the entry point."""
    return [
        'if __name__ == "__main__":',
        "    values: dict[str] = read_sql(DATABASE)",
        "    main(values)",
        "",
    ]


def print_error(**kwargs: dict) -> list[str]:
    """Help for print_fill_axe."""
    if kwargs["error"] is not None:
        return [
            "",
            "# compute error bar",
            'errs = list(set(sub_values["error"]))  # list to frozen order',
            "xerr, yerr = {e: [] for e in errs}, {e: [] for e in errs}",
            'for x, y, err in zip(x_data, y_data, sub_values["error"]):',
            "    xerr[err].append(x)",
            "    yerr[err].append(y)",
            "xerr = [xerr[e] for e in errs]",
            "yerr = [yerr[e] for e in errs]",
            "x_data, xerr = zip(*map(mean_std, xerr))",
            "y_data, yerr = zip(*map(mean_std, yerr))",
            "",
        ]
    return [
        "xerr = None",
        "yerr = None",
    ]


def print_cst(**kwargs: dict) -> list[str]:
    """Return the code of the definition of the constants."""
    return [
        f"COLORS = {COLORS}",
        f"DATABASE = pathlib.Path({str(kwargs['database'])!r})",
        'FIGNAME = pathlib.Path(__file__).with_suffix(".svg")',
        f"FIGSIZE = {FIGSIZE}  # (width, height)",
        f"MARKERS = {MARKERS}  # list of dot shapes",
        f"MARKERSIZE = {MARKERSIZE}",
        "",
        "",
    ]


def print_grid(axe: str, dim: str, scale: Scale) -> list[str]:
    """Define the ticks rule and log or linear scale.

    Parameters
    ----------
    axe : str
        The acessor to the matplotlib axe.
    dim : str
        "x" or "y".
    scale : Scale
        Linear, logarithmic or discrete.

    """
    assert isinstance(axe, str), axe.__class__.__name__
    assert isinstance(scale, Scale), scale.__class__.__name__
    assert dim in {"x", "y"}, dim

    if scale == Scale.DISCRETE:
        logging.getLogger(__name__).warning("the axis %s should be display as bar plot", dim)
        return []
    if scale == Scale.LOGARITHMIC:
        return [
            f'{axe}.set_{dim}scale("log", base=10)  # or .set_{dim}scale("linear")',
            f'{axe}.grid(axis="{dim}", which="major")',
            f'{axe}.grid(axis="{dim}", which="minor", color="0.9")',
        ]
    if scale == Scale.LINEAR:
        return [
            f'{axe}.set_{dim}scale("linear")  # or .set_{dim}scale("log", base=10)',
            f'{axe}.grid(axis="{dim}", which="major")',
            f'{axe}.grid(axis="{dim}", which="minor", color="0.9")',
            f"ticks = {axe}.{dim}axis.get_majorticklocs()  # get main graduations",
            f"{axe}.{dim}axis.set_minor_locator(ticker.MultipleLocator((ticks[1]-ticks[0])/10.0))",
        ]
    msg = f"{scale} is not valid"
    raise ValueError(msg)


def print_fill_axe(**kwargs: dict) -> list[str]:
    """Return the code that fill a given axes."""
    code = [
        "def fill_axe(axe: plt.Axes, values: dict[str], ylab: str, xlab: str):",
        '    """Fill in the axis with the data provided as input."""',
    ]
    if kwargs["color"] is not None:
        code.append('    color_labels = sorted(set(values["color"]), key=smartsort)')
    if kwargs["marker"] is not None:
        code.append('    marker_labels = sorted(set(values["marker"]), key=smartsort)')
    match (kwargs["color"] is None, kwargs["marker"] is None):
        case (True, True):  # no color, no marker
            code.extend([
                "    x_data, y_data = values[xlab], values[ylab]",
                *(f"    {line}" for line in print_error(**kwargs)),
                "    axe.errorbar(",
                "        x_data, y_data,",
                "        xerr=xerr, yerr=yerr,",
                "        color=COLORS[0],",
                "        fmt=MARKERS[0],",
                "        capsize=3,",
                "        markersize=MARKERSIZE,",
                "    )",
            ])
        case (False, True):  # only color
            code.extend([
                "    for color_label, color in zip(color_labels, itertools.cycle(COLORS)):",
                '        mask = [v == color_label for v in values["color"]]',
                "        sub_values = {",
                "            k: [v for v, m in zip(vs, mask) if m] for k, vs in values.items()",
                "        }",
                "        x_data, y_data = sub_values[xlab], sub_values[ylab]",
                *(f"        {line}" for line in print_error(**kwargs)),
                "        axe.errorbar(",
                "            x_data, y_data,",
                "            xerr=xerr, yerr=yerr,",
                f'            label=f"{kwargs["color"]}={{color_label}}", ',
                "            color=color,",
                "            fmt=MARKERS[0],",
                "            capsize=3,",
                "            markersize=MARKERSIZE",
                "        )",
            ])
        case (True, False):  # only marker
            code.extend([
                "    for marker_label, marker in zip(marker_labels, itertools.cycle(MARKERS)):",
                '        mask = [v == marker_label for v in values["marker"]]',
                "        sub_values = {",
                "            k: [v for v, m in zip(vs, mask) if m] for k, vs in values.items()",
                "        }",
                "        x_data, y_data = sub_values[xlab], sub_values[ylab]",
                *(f"        {line}" for line in print_error(**kwargs)),
                "        axe.errorbar(",
                "            x_data, y_data,",
                "            xerr=xerr, yerr=yerr,",
                f'            label=f"{kwargs["marker"]}={{marker_label}}", ',
                "            color=COLORS[0],",
                "            fmt=marker,",
                "            capsize=3,",
                "            markersize=MARKERSIZE",
                "        )",
            ])
        case (False, False):  # both color and marker
            code.extend([
                "",
                "    # create legend",
                "    for color_label, color in zip(color_labels, itertools.cycle(COLORS)):",
                (
                    "        axe.errorbar("
                    f'[], [], color=color, label=f"{kwargs["color"]}={{color_label}}"'
                    ")"
                ),
                "    for marker_label, marker in zip(marker_labels, itertools.cycle(MARKERS)):",
                (
                    "        axe.errorbar("
                    '[], [], color="grey", fmt=marker, '
                    f'label=f"{kwargs["marker"]}={{marker_label}}"'
                    ")"
                ),
                "",
                "    # plot data",
                "    for (color_label, color), (marker_label, marker) in itertools.product(",
                "        zip(color_labels, itertools.cycle(COLORS)),",
                "        zip(marker_labels, itertools.cycle(MARKERS)),",
                "    ):",
                "        mask = [",
                "            vc == color_label and vm == marker_label",
                '            for vc, vm in zip(values["color"], values["marker"])',
                "        ]",
                "        sub_values = {",
                "            k: [v for v, m in zip(vs, mask) if m] for k, vs in values.items()",
                "        }",
                "        x_data, y_data = sub_values[xlab], sub_values[ylab]",
                *(f"        {line}" for line in print_error(**kwargs)),
                "        axe.errorbar(",
                "            x_data, y_data,",
                "            xerr=xerr, yerr=yerr,",
                "            color=color,",
                "            fmt=marker,",
                "            capsize=3,",
                "            markersize=MARKERSIZE",
                "        )",
            ])

    code.extend(["", ""])
    return code


def print_header() -> list[str]:
    """Return the code at the very top of the file."""
    return [
        "#!/usr/bin/env python3",
        "",
        (
            f'"""Code auto-generated by mendevi {__version__} '
            f'the {datetime.datetime.now(tz=datetime.UTC)}."""'
        ),
        "",
    ]


def print_import() -> list[str]:
    """Print the importations."""
    return [
        "import itertools",
        "import math",
        "import pathlib",
        "import sqlite3",
        "",
        "from context_verbose import Printer",
        "from mendevi.exceptions import RejectError",
        "from mendevi.plot.tools import smartsort",
        "from mendevi.utils import best_profile",
        "import cutcutcodec",
        "import matplotlib.figure as figure",
        "import matplotlib.pyplot as plt",
        "import matplotlib.ticker as ticker",
        "import mendevi.database.extract as extract",
        "import numpy as np",
        "import tqdm",
        "",
        "",
    ]


def print_subfigure(**kwargs: dict) -> list[str]:
    """Return the code for the subplot of a a given y and y axis."""
    header = [
        (
            f"def plot_{safe_lbl(kwargs['ylabel'])}_{safe_lbl(kwargs['xlabel'])}("
            "values: dict[str], subfig: figure.SubFigure"
            "):"
        ),
        f'    """Create the subchart for y={kwargs["ylabel"]} and x={kwargs["xlabel"]}."""',
    ]
    xlabel = f"x_{kwargs['x'].index(kwargs['xlabel'])}"
    ylabel = f"y_{kwargs['y'].index(kwargs['ylabel'])}"
    scale_x = get_extractor(kwargs["xlabel"], safe=True).scale
    scale_y = get_extractor(kwargs["ylabel"], safe=True).scale
    match (kwargs["window_y"] is None, kwargs["window_x"] is None):
        case (True, True):
            middle = [
                "    # create a simple figure",
                '    axe = subfig.subplots(sharex="col")  # or "all"',
                f"    fill_axe(axe, values, ylab={ylabel!r}, xlab={xlabel!r})",
                *(f"    {line}" for line in print_grid("axe", "y", scale=scale_y)),
                *(f"    {line}" for line in print_grid("axe", "x", scale=scale_x)),
            ]
        case (False, True):
            middle = [
                "    # create a 1d vertical subplot figure",
                '    all_window_y_values = sorted(set(values["window_y"]), key=smartsort)',
                "    nrows = len(all_window_y_values)",
                '    axes = subfig.subplots(nrows=nrows, squeeze=False, sharex="col")  # or "all"',
                "",
                "    # iterate over all subplots",
                "    for i, window_y_value in enumerate(all_window_y_values):",
                "        # select the correct data subset",
                '        mask = [y == window_y_value for y in values["window_y"]]',
                "        sub_values = {",
                "            k: [v for v, m in zip(vs, mask) if m] for k, vs in values.items()",
                "        }",
                f'        axes[i, 0].set_title({kwargs["window_y"]!r} f"={{window_y_value}}")',
                f"        fill_axe(axes[i, 0], sub_values, ylab={ylabel!r}, xlab={xlabel!r})",
                *(f"        {line}" for line in print_grid("axes[i, 0]", "y", scale=scale_y)),
                *(f"        {line}" for line in print_grid("axes[i, 0]", "x", scale=scale_x)),
            ]
        case (True, False):
            middle = [
                "    # create a 1d horizontal subplot figure",
                '    all_window_x_values = sorted(set(values["window_x"]), key=smartsort)',
                "    ncols = len(all_window_x_values)",
                '    axes = subfig.subplots(ncols=ncols, squeeze=False, sharey="row")  # or "all"',
                "",
                "    # iterate over all subplots",
                "    for j, window_x_value in enumerate(all_window_x_values):",
                "        # select the correct data subset",
                '        mask = [x == window_x_value for x in values["window_x"]]',
                "        sub_values = {",
                "            k: [v for v, m in zip(vs, mask) if m] for k, vs in values.items()",
                "        }",
                f'        axes[0, j].set_title({kwargs["window_x"]!r} f"={{window_x_value}}")',
                f"        fill_axe(axes[0, j], sub_values, ylab={ylabel!r}, xlab={xlabel!r})",
                *(f"        {line}" for line in print_grid("axes[0, j]", "y", scale=scale_y)),
                *(f"        {line}" for line in print_grid("axes[0, j]", "x", scale=scale_x)),
            ]
        case (False, False):  # 2d grid subplot
            middle = [
                "    # create a 2d grid subplot figure",
                '    all_window_y_values = sorted(set(values["window_y"]), key=smartsort)',
                '    all_window_x_values = sorted(set(values["window_x"]), key=smartsort)',
                "    nrows, ncols = len(all_window_y_values), len(all_window_x_values)",
                "    axes = subfig.subplots(",
                '        nrows, ncols, squeeze=False, sharex="col", sharey="row",',
                '    )  # or "all"',
                "",
                "    # iterate over all subplots",
                "    for (i, window_y_value), (j, window_x_value) in itertools.product(",
                "        enumerate(all_window_y_values), enumerate(all_window_x_values)",
                "    ):",
                "        # select the correct data subset",
                "        mask = [",
                "            y == window_y_value and x == window_x_value",
                '            for y, x in zip(values["window_y"], values["window_x"])',
                "        ]",
                "        sub_values = {",
                "            k: [v for v, m in zip(vs, mask) if m] for k, vs in values.items()",
                "        }",
                "        axes[i, j].set_title(",
                f'            {kwargs["window_y"]!r} f"={{window_y_value}}"',
                '            " & "',
                f'            {kwargs["window_x"]!r} f"={{window_x_value}}"',
                "        )",
                f"        fill_axe(axes[i, j], sub_values, ylab={ylabel!r}, xlab={xlabel!r})",
                *(f"        {line}" for line in print_grid("axes[i, j]", "y", scale=scale_y)),
                *(f"        {line}" for line in print_grid("axes[i, j]", "x", scale=scale_x)),
            ]
    end = [
        f"    subfig.supylabel({get_extractor(kwargs['ylabel'], safe=True).legend!r})",
        f"    subfig.supxlabel({get_extractor(kwargs['xlabel'], safe=True).legend!r})",
        "",
    ]
    return header + middle + end


def print_main(**kwargs: dict) -> list[str]:
    """Return the code of the function that plot the final graphic."""
    code = []
    # subplots
    for ylabel, xlabel in itertools.product(kwargs["y"], kwargs["x"]):
        code.extend(print_subfigure(**kwargs, ylabel=ylabel, xlabel=xlabel))

    # main function
    code.extend([
        "def main(values: dict[str]):",
        '    """Create the plots with matplotlib."""',
        "    # create and fill in the main figure",
        '    fig = plt.figure(layout="constrained", figsize=FIGSIZE)',
        (
            "    subfigs = fig.subfigures("
            f"nrows={len(kwargs['y'])}, ncols={len(kwargs['x'])}, squeeze=False"
            ")"
        ),
    ])
    for (i, ylabel), (j, xlabel) in itertools.product(
        enumerate(kwargs["y"]), enumerate(kwargs["x"]),
    ):
        code.extend([
            f"    plot_{safe_lbl(ylabel)}_{safe_lbl(xlabel)}(values, subfigs[{i}, {j}])",
        ])
    code.extend([
        "",
        "    axes = [a for f in subfigs.flat for a in np.ravel(f.axes)]",
        "",
        "    # adjusts the figsize according to the number of graphs",
        "    nrows = np.vectorize(",
        "        lambda s: s.axes[0].get_gridspec().nrows",
        "    )(subfigs).sum(axis=0).max()",
        "    ncols = np.vectorize(",
        "        lambda s: s.axes[0].get_gridspec().ncols",
        "    )(subfigs).sum(axis=1).max()",
        "    fig.set_size_inches(FIGSIZE[0]*math.sqrt(ncols), FIGSIZE[1]*math.sqrt(nrows))",
        "",
        "    # recompute limits to solve share log errorbar matplotlib error",
        "    for axe in axes:",
        "        axe.relim()  # recompute limits",
        "        axe.autoscale()  # adjust the view",
        "",
        "    # legend managment",
        "    labels = {frozenset(a.get_legend_handles_labels()[1]) for a in axes}",
        "    if len(axes) != 1 and len(labels) == 1:  # same legend everywhere",
        "        lines, labels = axes[0].get_legend_handles_labels()",
        "        if labels:",
        "            fig.legend(",
        '                lines, labels, loc="outside upper center", ncols=min(5, len(labels))',
        "            )",
        "    else:",
        "        for subfig in subfigs.flat:",
        "            axes = np.ravel(subfig.axes)",
        "            labels = {frozenset(a.get_legend_handles_labels()[1]) for a in axes}",
        "            if len(axes) != 1 and len(labels) == 1:  # same legend in all axes",
        "                lines, labels = axes[0].get_legend_handles_labels()",
        "                if labels:",
        "                    subfig.legend(",
        (
            "                        "
            'lines, labels, loc="outside upper center", ncols=min(3, len(labels))'
        ),
        "                    )",
        "            else:  # legends are different for each axe of that subfigure",
        "                for axe in axes:",
        "                    axe.legend()",
        "",
    ])
    keys = ["x", "y", "color", "marker", "window_x", "window_y", "error", "filter", "table"]
    code.extend([
        "",
        "    # save the figure",
        "    metadata = {",
        '        "Keywords": ["mendevi", "video", "encoding", "decoding"],',
        '        "Language": "en",',
        '        "Rights": "Free lisence GLPv3",',
        '        "Contributor": ["Robin RICHARD (robinechuca)"],',
        f'        "Creator": "mendevi {__version__}",',
        f'        "Description": "A plot with {", ".join(f"{k}={kwargs[k]}" for k in keys)}",',
        f'        "Title": {kwargs["path"].stem!r},',
        "    }",
        "    plt.savefig(FIGNAME, format=FIGNAME.suffix[1:], transparent=False, metadata=metadata)",
        "    plt.show()",
        "",
        "",
    ])
    return code


def print_mean_std() -> list[str]:
    """Return the code to compute the mean and the std of a list."""
    return [
        "def mean_std(data: list[float]) -> tuple[float, float]:",
        '    """Compute the mean and the unbiaised std of a list."""',
        "    data = [d for d in data if d is not None]",
        "    match nbr := float(len(data)):",
        "        case 0.0:",
        "            return math.nan, math.nan",
        "        case 1.0:",
        "           return data[0], math.nan",
        "    mean = sum(data) / nbr",
        "    if (var_sum := sum((x-mean)**2 for x in data)) == 0.0:",
        "        return mean, math.nan",
        "    std = math.sqrt(var_sum / (nbr - 1.0))",
        "    return mean, std",
        "",
        "",
    ]


def print_read_sql(**kwargs: dict) -> list[str]:
    """Return the code of the function that perform the sql request."""
    alias = {
        f"x_{i}": expr for i, expr in enumerate(kwargs["x"])
    } | {
        f"y_{i}": expr for i, expr in enumerate(kwargs["y"])
    } | {
        lbl: kwargs[lbl]
        for lbl in ["color", "marker", "window_x", "window_y", "error"]
        if kwargs[lbl] is not None
    }

    # get sql query
    atom_names, line_extractor = merge_extractors(set(alias), alias=alias, select=kwargs["filter"])
    select = {s for lbl in atom_names for s in get_extractor(lbl).func.select}
    if len(queries := SqlLinker(*select).sql) == 0:
        logging.getLogger(__name__).warning(
            "fail to create the SQL query, please provide it yourself",
        )
        queries = [""]

    # select good query
    if kwargs["table"] is not None:
        table = kwargs["table"]
        queries = {re.search(r"FROM\s+(?P<tab>\w+)", q)["tab"]: q for q in queries}
        if table not in queries:
            msg = f"possible queries from {', '.join(queries)}, not {table}"
            raise ValueError(msg)
        queries = [queries[table]]
    else:
        table = re.search(r"FROM\s+(?P<tab>\w+)", queries[0])["tab"]

    # include in template
    return [
        *line_extractor,
        "",
        "",
        "def thread_line_extractor(row: sqlite3.Row) -> dict[str] | None:",
        '    """Multiprocessing compatible version of `line_extractor`."""',
        "    assert isinstance(row, sqlite3.Row), row.__class__.__name__",
        "    row = dict(row)",
        "    try:",
        "        return line_extractor(row)",
        "    except RejectError:",
        "        return None",
        "",
        "",
        "def read_sql(database: pathlib.Path) -> dict[str, list]:",
        '    """Extract the relevant values from the database."""',
        f"    values: dict[str, list] = {{{', '.join(f'{n!r}: []' for n in sorted(alias))}}}",
        (
            '    with sqlite3.connect(f"file:{database}?mode=ro", uri=True) as conn, '
            'Printer("Read database...") as prt:'
        ),
        f'        (count,) = conn.execute("SELECT COUNT(*) FROM {table}").fetchone()',
        "        conn.row_factory = sqlite3.Row",
        "        for line in tqdm.tqdm(",
        "            cutcutcodec.core.opti.parallel.imap(",
        "                thread_line_extractor,",
        "                conn.execute(",
        '                    """',
        f"                    {'\n                    '.join(queries[0].split('\n'))}",
        '                    """',
        *[
            f"                    # {q}"
            for q in "\n".join(f'"""\n{q}\n"""' for q in queries[1:]).split("\n")
        ],
        "                ),",
        "            ),",
        "            dynamic_ncols=True,",
        "            leave=False,",
        "            smoothing=1e-6,",
        "            total=count,",
        '            unit="line",',
        "        ):",
        "            if line is not None:",
        "                for label, value in line.items():",
        "                    values[label].append(value)",
        "        prt.print_time()",
        "    return values",
        "",
        "",
    ]


def safe_lbl(lbl: str) -> str:
    """Convert the label into a valid lowercase variable."""
    assert isinstance(lbl, str), lbl.__class__.__name__
    for symb, rep in [("+", "plus"), ("-", "minus"), ("*", "times"), ("/", "per")]:
        lbl = lbl.replace(symb, f"_{rep}_")
    return re.sub(r"[^0-9a-zA-Z]+", "_", lbl.lower())



"""Management of all unit tests."""

import click

from mendevi.testing.run import run_tests


@click.command()
@click.option("--debug", is_flag=True, help="Show more information on failure.")
@click.option("--skip-install", is_flag=True, help="Do not check the installation.")
@click.option("--skip-coding-style", is_flag=True, help="Do not check programation style.")
@click.option("--skip-slow", is_flag=True, help="Do not run the slow tests.")
def main(*, debug: bool = False, **kwargs: dict) -> int:
    """Run several tests."""
    return run_tests(
        debug=debug,
        skip_install=kwargs.get("skip_install", False),
        skip_coding_style=kwargs.get("skip_coding_style", False),
        skip_slow=kwargs.get("skip_slow", False),
    )



"""The constants for graphical plot style."""

COLORS = [
    "#ef597b",  # bright pink
    "#29a2c6",  # cyan blue
    "#ffcb18",  # vivid yellow
    "#73b66b",  # medium green
    "#4e4f4b",  # dark gray
    "#a23e9a",  # purple
    "#f59331",  # orange
    "#007f5f",  # deep teal green
    "#00509e",  # deep blue
    "#c1b7b0",  # light beige gray
    "#ff6f00",  # strong orange
    "#0089a3",  # ocean blue
    "#b2df8a",  # light lime green
    "#cab2d6",  # soft lavender
    "#6a3d9a",  # dark violet
    "#fb9a99",  # soft pink
    "#1b9e77",  # forest green
    "#d95f02",  # burnt orange
    "#7570b3",  # indigo
    "#e7298a",  # magenta
]
FIGSIZE = (10.0, 8.0)  # (width, height)
MARKERS = [
    "o",
    "x",
    "+",
    "^",
    "d",
    "s",
    "v",
    "*",
    "D",
    "H",
    "p",
    "<",
    ">",
    "1",
    "2",
]  # list of dot shapes
MARKERSIZE = 4



"""Perform encoding measures."""

import fractions
import hashlib
import itertools
import sqlite3
from typing import TYPE_CHECKING

import click
import numpy as np
from context_verbose import Printer

from mendevi.cst.encoders import ENCODERS
from mendevi.database.complete import add_environment
from mendevi.encode import encode_and_store
from mendevi.utils import compute_video_hash, get_pix_fmt, get_rate_video, get_resolution

from .parse import PixelParamType, ResolutionParamType, parse_videos_database

if TYPE_CHECKING:
    import pathlib


def _parse_args(prt: Printer, kwargs: dict) -> None:
    """Verification of the arguments."""
    # repeat
    assert isinstance(kwargs["repeat"], int), kwargs["repeat"].__class__.__name__
    assert kwargs["repeat"] >= 1, kwargs["repeat"]
    prt.print(f"repeat    : {kwargs['repeat']}")

    # effort
    assert isinstance(kwargs["effort"], tuple), kwargs["effort"].__class__.__name__
    assert all(isinstance(p, str) for p in kwargs["effort"]), kwargs["effort"]
    assert all(p in {"fast", "medium", "slow"} for p in kwargs["effort"]), kwargs["effort"]
    prt.print(f"efforts   : {', '.join(kwargs['effort'])}")

    # encoder
    assert isinstance(kwargs["encoder"], tuple), kwargs["encoder"].__class__.__name__
    assert all(isinstance(e, str) for e in kwargs["encoder"]), kwargs["encoder"]
    assert all(e in ENCODERS for e in kwargs["encoder"]), kwargs["encoder"]
    prt.print(f"encoders  : {', '.join(kwargs['encoder'])}")

    # points
    assert isinstance(kwargs["points"], int), kwargs["points"].__class__.__name__
    assert kwargs["points"] >= 1, kwargs["points"]
    kwargs["quality"] = np.linspace(
        1.0/(kwargs["points"]+1),
        kwargs["points"]/(kwargs["points"]+1),
        kwargs["points"],
        dtype=np.float16,
    ).tolist()
    prt.print(f"qualities : k/{kwargs['points']+1}, k \u2208 [1, {kwargs['points']}]")

    # threads
    assert isinstance(kwargs["threads"], tuple), kwargs["threads"].__class__.__name__
    assert all(isinstance(t, int) for t in kwargs["threads"]), kwargs["threads"]
    assert all(t >= 1 for t in kwargs["threads"]), kwargs["threads"]
    prt.print(f"threads   : {', '.join(map(str, kwargs['threads']))}")

    # fps
    assert isinstance(kwargs["fps"], tuple), kwargs["fps"].__class__.__name__
    assert all(isinstance(f, fractions.Fraction) for f in kwargs["fps"]), kwargs["fps"]
    assert all(f > 0 for f in kwargs["fps"]), kwargs["fps"]
    if kwargs["fps"]:
        kwargs["fps"] = [f.limit_denominator(1001) for f in kwargs["fps"]]
        prt.print(f"fps       : {', '.join(map(str, kwargs['fps']))}")
    else:
        kwargs["fps"] = (None,)

    # resolution
    assert isinstance(kwargs["resolution"], tuple), kwargs["resolution"].__class__.__name__
    assert all(
        isinstance(r, tuple) and len(r) == 2 and all(isinstance(s, int) and s > 0 for s in r)
        for r in kwargs["resolution"]
    ), kwargs["resolution"]
    if kwargs["resolution"]:
        prt.print(f"resolution: {', '.join(f'w={w}:h={h}' for h, w in kwargs['resolution'])}")
    else:
        kwargs["resolution"] = (None,)

    # pix_fmt
    assert isinstance(kwargs["pix_fmt"], tuple), kwargs["pix_fmt"].__class__.__name__
    assert all(isinstance(p, str) for p in kwargs["pix_fmt"]), kwargs["pix_fmt"]
    if kwargs["pix_fmt"]:
        prt.print(f"pix_fmt   : {', '.join(kwargs['pix_fmt'])}")
    else:
        kwargs["pix_fmt"] = (None,)

    # mode
    assert isinstance(kwargs["mode"], tuple), kwargs["mode"].__class__.__name__
    assert all(isinstance(p, str) and p in {"cbr", "vbr"} for p in kwargs["mode"]), kwargs["mode"]
    kwargs["mode"] = tuple(sorted(set(kwargs["mode"])))
    prt.print(
        "br mode   : "
        f"{' and '.join({'vbr': 'variable', 'cbr': 'constant'}[m] for m in kwargs['mode'])}",
    )

    # filter
    assert isinstance(kwargs["filter"], tuple), kwargs["filter"].__class__.__name__
    assert all(isinstance(f, str) for f in kwargs["filter"]), kwargs["filter"]
    if kwargs["filter"]:
        prt.print(f"filters   : {', '.join(map(str, kwargs['filter']))}")
    else:
        kwargs["filter"] = (None,)


@click.command()
@click.argument("videos", type=click.Path(), nargs=-1)
@click.option("-d", "--database", type=click.Path(), help="The database path.")
@click.option(
    "-r", "--repeat",
    type=int,
    default=2,
    help="The number of times the encoding is repeated on this machine.",
)
@click.option(
    "-e", "--effort",
    type=click.Choice(["fast", "medium", "slow"]),
    default=["medium"],
    multiple=True,
    help="The compression effort (default = medium).",
)
@click.option(
    "-c", "--encoder",
    type=click.Choice(sorted(ENCODERS)),
    default=["libx264", "libsvtav1"],
    multiple=True,
    help="The encoder name.",
)
@click.option(
    "-n", "--points",
    type=int,
    default=16,
    help="The number of quality point per encoder.",
)
@click.option(
    "-t", "--threads",
    type=int,
    default=[8],
    multiple=True,
    help="The number of threads used by encoders.",
)
@click.option(
    "-f", "--fps",
    type=fractions.Fraction,
    multiple=True,
    help="The optional framerate conversion during encoding.",
)
@click.option(
    "--resolution",
    type=ResolutionParamType(),
    multiple=True,
    help="The optional video shape conversion during encoding.",
)
@click.option(
    "--pix_fmt",
    type=PixelParamType(),
    multiple=True,
    help="The optional pixel format conversion during encoding.",
)
@click.option(
    "-m", "--mode",
    type=click.Choice(["cbr", "vbr"]),
    default=["vbr"],
    multiple=True,
    help="The optional bitrate mode (cbr or vbr).",
)
@click.option(
    "--filter",
    type=str,
    multiple=True,
    help="The ffmpeg filter to apply before encoding.",
)
def main(videos: tuple, database: str | None = None, **kwargs: dict) -> None:
    """Measures activity during encoding.

    \b
    Parameters
    ----------
    videos : tuple[pathlike]
        The source videos to be transcoded.
        It can be a glob expression, a directory or a file path.
    database : pathlike, optional
        The path to the database where all measurements are stored.
        If a folder is provided, the database is created inside this folder.
        By default, it is created right next to the video.
    **kwargs: dict
        Please refer to the detailed arguments below.
    repeat : int, default=2
        The number of times the experiment is repeated on this environment.
        This allows us to estimate the variance of measurements.
    effort : tuple[str], default=("medium",)
        The effort made to compress, `fast`, `medium` or `slow`.
    encoder : tuple[str], default=("libx264", "libsvtav1")
        The encoders and therefore the codecs to use.
        The available encoders are mendevi.cst.encoders.ENCODERS.
    points : int, default=16
        The number of different qualities to use.
        It is an indirect way to determine the CRF or the QP.
        The quality values are distributed evenly over ]0, 1[,
        for example, points=3 => qualities=[0.25, 0.5, 0.75].
    threads : int, default=8
        The theoretical number of threads used by the encoder.
        This roughly reflects the number of logical cores used.
    fps : fractions.Fraction, optional
        If provided, the reference video will be resampled on fly just before being encoded.
        This is therefore the frame rate of the transcoded video.
        If this argument is not provided, the frame rate remains unchanged.
    resolution : tuple[int, int], optional
        If provided, the reference video will be reshaped on fly just before being encoded.
        This is therefore the resolution of the transcoded video.
        If this argument is not provided, the resolution remains unchanged.
    pix_fmt : str, optional
        If provided, A pixel conversion is performed during encoding.
    mode : tuple[str]
        Text that is ``cbr`` if the bit rate is constant (cbr mode),
        or ``vbr`` if it is variable (constant quality mode).
    filter : str, optional
        A video ffmpeg filter (after -vf) which applies at the time of transcoding,
        more precisely between decoding and conversion. It is a kind of pre-conversion.

    """
    with Printer("Parse configuration...") as prt:
        videos, database = parse_videos_database(prt, videos, database)
        assert videos, "at least one video is required"
        _parse_args(prt, kwargs)

    # preparation of context
    env_id = add_environment(database)
    kwargs["src_vid_id"]: dict[pathlib.Path, bytes] = compute_video_hash(videos)

    keys = [
        "effort", "encoder", "filter", "fps", "mode", "pix_fmt", "quality", "resolution", "threads",
    ]

    # retrieves the settings for videos that have already been transcoded
    with sqlite3.connect(database) as conn, Printer("Search already done...") as prt:
        conn.row_factory = sqlite3.Row
        done: dict[tuple, int] = {}
        for row in conn.execute(
            "SELECT * FROM t_enc_encode WHERE enc_env_id=?", (env_id,),
        ):
            values = {
                "effort": row["enc_effort"],
                "encoder": row["enc_encoder"],
                "filter": row["enc_filter"] or None,
                "fps": fractions.Fraction(row["enc_fps"]).limit_denominator(1001),
                "mode": row["enc_mode"],
                "pix_fmt": row["enc_pix_fmt"],
                "quality": float(row["enc_quality"]),
                "resolution": (row["enc_height"], row["enc_width"]),
                "src_vid_id": row["enc_src_vid_id"],
                "threads": row["enc_threads"],
            }
            values = tuple(values[k] for k in ["src_vid_id", *keys])
            done[values] = done.get(values, 0) + 1
        prt.print(f"{sum(done.values())} video already encoded under these environment")

    # iterate on all parameters
    loops = sorted(
        itertools.product(videos, *(kwargs[k] for k in keys)),
        key=lambda t: hashlib.md5(str(t).encode("utf-8")).hexdigest(),  # repetable shuffle
    )
    for i, (repeat, values_) in enumerate(itertools.product(range(kwargs["repeat"]), loops)):
        values = dict(zip(["video", *keys], values_, strict=True), repeat=values_[0])
        values["repeat"] = repeat
        values["fps"] = values["fps"] or get_rate_video(values["video"])
        values["pix_fmt"] = values["pix_fmt"] or get_pix_fmt(values["video"])
        values["resolution"] = values["resolution"] or get_resolution(values["video"])
        values["src_vid_id"] = kwargs["src_vid_id"][values["video"]]
        key = (values["src_vid_id"], *tuple(values[k] for k in keys))
        if done.get(key, 0) > values["repeat"]:
            continue
        with Printer(f"Encode {i+1}/{kwargs["repeat"]*len(loops)}...", color="cyan") as prt:
            encode_and_store(database, env_id, values.pop("video"), **values)
            done[key] = done.get(key, 0) + 1
            prt.print_time()



"""Automatate the decapsulation process."""

import pathlib
import re

from .decompress import decompress
from .torrent_finder import get_torrent
from .transmission import add_torrent


def retrive_file(file: pathlib.Path | str) -> pathlib.Path:
    """Attempt to retrieve the file.

    Parameters
    ----------
    file : pathlike
        The name or the identifier of the file you want to retrive.

    Returns
    -------
    decapsulated_file : pathlib.Path
        The final unencapsulated file.

    """
    file = pathlib.Path(file).expanduser()

    # handle <...>
    if re.match(r"^<.+>$", file.name):
        name = re.sub(r"^<(.+?)(?:\.xz(?:\.torrent)?)?>$", r"\1", file.name) + ".xz.torrent"
        return retrive_file(file.parent / name)

    # remove .xz.torrent
    file = file.parent / re.sub(r"\.xz(?:\.torrent)?$", "", file.name)

    # shortcut
    if file.exists():
        return file

    # handle .xz
    if (prev_file := file.parent / f"{file.name}.xz").exists():
        return decompress(prev_file)

    # handle .torrent
    if (prev_prev_file := prev_file.parent / f"{prev_file.name}.torrent").exists():
        return retrive_file(add_torrent(prev_prev_file))

    # general case
    return retrive_file(get_torrent(prev_prev_file.name))



"""Executes all the tests via the ``pytest`` module."""

try:
    import pytest
except ImportError as err:
    msg = "pytest paquage required (uv pip install mendevi[test])"
    raise ImportError(msg) from err

from mendevi.utils import get_project_root


def run_tests(
    *,
    debug: bool = False,
    skip_install: bool = False,
    skip_coding_style: bool = False,
    skip_slow: bool = False,
) -> int:
    """Perform all unit tests."""
    root = get_project_root()
    debug_options = ["--full-trace"] if debug else []

    # install check
    if not skip_install:
        print("Checking if the dependencies are corrected...")
        path = str(root / "testing" / "install.py")
        if (
            code := pytest.main([*debug_options, "--verbose", path])
        ):
            return int(code)

    # code quality check
    if not skip_coding_style:
        print("Checking if the coding style respects the PEP...")
        path = str(root / "testing" / "coding_style.py")
        if (
            code := pytest.main([
                *debug_options,
                "--verbose", "--exitfirst", "--capture=no", "--tb=no", "-rN", path,  # no rapport
            ])
        ):
            return int(code)

    # classical tests
    print("Runing all the little unit tests...")
    paths = sorted(
        map(
            str,
            (
                set(root.rglob("*.py"))
                - {root / "testing" / "install.py", root / "testing" / "coding_style.py"}
            ),
        ),
    )
    if (code := pytest.main([*debug_options, "-m", "not slow", "--doctest-modules", *paths])):
        return int(code)

    # slow tests
    if not skip_slow:
        print("Runing the slow unit tests...")
        if (code := pytest.main([*debug_options, "-m", "slow", "--verbose", *paths])):
            return int(code)
    return 0



"""Get the information of the environment."""

import logging
import os
import platform
import re
import subprocess

import psutil


def get_ffmpeg_version() -> str:
    r"""Return the version of ffmpeg.

    Examples
    --------
    >>> import pprint
    >>> from mendevi.measures.context import get_ffmpeg_version
    >>> pprint.pprint(get_ffmpeg_version())  # doctest: +ELLIPSIS
    ('ffmpeg version ... Copyright (c) '
     '... the FFmpeg developers\n'
     'built with gcc ...\n'
     'configuration: --prefix=.../ffmpeg_build '
     '--pkg-config-flags=--static '
     '--extra-cflags=-I/home/rrichard/ffmpeg_build/include '
     "--extra-ldflags=-L/home/rrichard/ffmpeg_build/lib --extra-libs='-lpthread "
     "-lm' --ld=g++ --bindir=/home/rrichard/bin --enable-chromaprint "
     '--enable-frei0r --enable-gpl --enable-ladspa --enable-libaom --enable-libass '
     '--enable-libbluray --enable-libbs2b --enable-libcaca --enable-libcdio '
     '--enable-libdav1d --enable-libdrm --enable-libfdk-aac --enable-libflite '
     '--enable-libfontconfig --enable-libfreetype --enable-libfribidi '
     '--enable-libgme --enable-libgsm --enable-libharfbuzz --enable-libmp3lame '
     '--enable-libopenmpt --enable-libopus --enable-libpulse --enable-librav1e '
     '--enable-librubberband --enable-librubberband --enable-libshine '
     '--enable-libsnappy --enable-libsoxr --enable-libspeex --enable-libsvtav1 '
     '--enable-libtheora --enable-libtwolame --enable-libvidstab '
     '--enable-libvo-amrwbenc --enable-libvorbis --enable-libvpl --enable-libvpx '
     '--enable-libvvenc --enable-libwebp --enable-libx264 --enable-libx265 '
     '--enable-libxml2 --enable-libxvid --enable-libzimg --enable-libzvbi '
     '--enable-nonfree --enable-opengl --enable-openssl --enable-vaapi '
     '--enable-version3\n'
     'libavutil ...\n'
     'libavcodec ...\n'
     'libavformat ...\n'
     'libavdevice ...\n'
     'libavfilter ...\n'
     'libswscale ...\n'
     'libswresample ...')
    >>>

    """
    out = subprocess.run(
        ["ffmpeg", "-version"], check=False, capture_output=True,
    ).stdout.decode()
    lines = out.split("\n")
    lines = [line for line in lines if line not in {"", "Exiting with exit code 0"}]
    return "\n".join(lines)


def get_libx265_version() -> str:
    """Return the version of the libx265 encoder.

    Examples
    --------
    >>> from mendevi.measures.context import get_libx265_version
    >>> print(get_libx265_version())  # doctest: +ELLIPSIS
    4...
    [Linux][GCC ...][64 bit] 8...
    MMX2 ... AVX2
    >>>

    """
    out = subprocess.run(
        [
            "ffmpeg", "-f", "lavfi", "-i", "nullsrc", "-frames:v", "1",
            "-c:v", "libx265",
            "-f", "null", "-",
        ],
        check=False, capture_output=True,
    ).stderr.decode()

    # version
    if (match := re.search(r"HEVC encoder version (?P<version>\S+)", out)) is None:
        logging.getLogger(__name__).warning("failed to find the libx265 version")
        version = ""
    else:
        version = match["version"]
    # build
    if (match := re.search(r"build info (?P<build>.+)\n", out)) is None:
        logging.getLogger(__name__).warning("failed to find the libx265 build info")
        build = ""
    else:
        build = match["build"]
    # cpu
    if (match := re.search(r"using cpu capabilities: (?P<cpu>.+)\n", out)) is None:
        logging.getLogger(__name__).warning("failed to find the libx265 build info")
        cpu = ""
    else:
        cpu = match["cpu"]

    return f"{version}\n{build}\n{cpu}"


def get_libvpx_vp9_version() -> str:
    """Return the version of the libvpx-vp9 encoder.

    Examples
    --------
    >>> from mendevi.measures.context import get_libvpx_vp9_version
    >>> print(get_libvpx_vp9_version())  # doctest: +ELLIPSIS
    v...
    >>>

    """
    out = subprocess.run(
        [
            "ffmpeg", "-f", "lavfi", "-i", "nullsrc", "-frames:v", "1",
            "-c:v", "libvpx-vp9",
            "-f", "null", "-",
        ],
        check=False, capture_output=True,
    ).stderr.decode()
    if (match := re.search(r"\[libvpx-vp9 @ \w+\] (?P<version>\S+)", out)) is None:
        logging.getLogger(__name__).warning("failed to find the libvpx-vp9 version")
        return ""
    return match["version"]


def get_libsvtav1_version() -> str:
    """Return the version of the libsvtav1 encoder.

    Examples
    --------
    >>> from mendevi.measures.context import get_libsvtav1_version
    >>> print(get_libsvtav1_version())  # doctest: +ELLIPSIS
    ...
    GCC ... 64 bit
    >>>

    """
    out = subprocess.run(
        [
            "ffmpeg", "-f", "lavfi", "-i", "nullsrc", "-frames:v", "1",
            "-c:v", "libsvtav1",
            "-f", "null", "-",
        ],
        check=False, capture_output=True,
    ).stderr.decode()

    # version
    if (match := re.search(r"SVT-AV1 Encoder Lib (?P<version>\S+)", out)) is None:
        logging.getLogger(__name__).warning("failed to find the libsvtav1 version")
        version = ""
    else:
        version = match["version"]
    # build
    if (match := re.search(r"SVT \[build\]\s*:\s*(?P<build>\S.*\S)\s*\n", out)) is None:
        logging.getLogger(__name__).warning("failed to find the libsvtav1 build info")
        build = ""
    else:
        build = re.sub(r"\s+", " ", match["build"])

    return f"{version}\n{build}"


def get_librav1e_version() -> str:
    """Return the version of the librav1e encoder.

    Examples
    --------
    >>> from mendevi.measures.context import get_librav1e_version
    >>> print(get_librav1e_version())  # doctest: +ELLIPSIS
    rav1e 0.8.0 (...) (release)
    rustc 1.91.1 (...) x86_64-unknown-linux-gnu
    Compiled CPU Features: adx,aes,...,xsaveopt,xsaves
    Runtime Assembly Support: Enabled
    Runtime Assembly Level: ...
    Threading: Enabled
    Unstable Features: Disabled
    Compiler Flags: -C target-cpu=native
    >>>

    """
    try:
        return subprocess.run(
            ["rav1e", "--version"],
            check=False, capture_output=True,
        ).stdout.decode().strip()
    except FileNotFoundError:
        logging.getLogger(__name__).warning("failed to find the librav1e version")
        return ""


def get_vvc_version() -> str:
    """Return the version of the vvc encoder.

    Examples
    --------
    >>> from mendevi.measures.context import get_vvc_version
    >>> print(get_vvc_version())
    1.13.1
    >>>

    """
    out = subprocess.run(
        [
            "ffmpeg", "-f", "lavfi", "-i", "nullsrc", "-frames:v", "1",
            "-c:v", "vvc",
            "-f", "null", "-",
        ],
        check=False, capture_output=True,
    ).stderr.decode()
    if (match := re.search(r"libvvenc version: (?P<version>\S+)", out)) is None:
        logging.getLogger(__name__).warning("failed to find the vvc version")
        return ""
    return match["version"]


def get_platform() -> dict:
    """Get basic information about the system.

    Examples
    --------
    >>> import pprint
    >>> from mendevi.measures.context import get_platform
    >>> pprint.pprint(get_platform())  # doctest: +ELLIPSIS
    {'hostname': '...',
     'kernel_version': 'Linux-...',
     'logical_cores': ...,
     'physical_cores': ...,
     'processor': '...',
     'python_compiler': '...',
     'python_version': '...',
     'ram': ...,
     'swap': ...,
     'system_version': '...'}
    >>>

    """
    return {
        "hostname": platform.node(),
        "kernel_version": platform.platform(),
        "logical_cores": psutil.cpu_count(logical=True),
        "physical_cores": psutil.cpu_count(logical=False),
        "processor": platform.processor() or platform.machine(),
        "python_compiler": platform.python_compiler(),
        "python_version": platform.python_version(),
        "ram": psutil.virtual_memory().total,
        "swap": psutil.swap_memory().total,
        "system_version": platform.version(),
    }


def get_pip_freeze() -> str:
    """Return the sorted pip freeze.

    Examples
    --------
    >>> from mendevi.measures.context import get_pip_freeze
    >>> print(get_pip_freeze())  # doctest: +ELLIPSIS
    accessible-pygments==...
    ...
    wheel==...
    >>>

    """
    return subprocess.run(
        ["pip", "freeze"], check=False, capture_output=True,
    ).stdout.decode().rstrip()


def get_lshw() -> str:
    """Extract the very accurate exaustive info as big json.

    Examples
    --------
    >>> from mendevi.measures.context import get_lshw
    >>> full_info = get_lshw()
    >>>

    """
    if os.geteuid() != 0:
        logging.getLogger(__name__).warning("you should run as super user to get more system info")
    try:
        out = subprocess.run(
            ["lshw", "-json"], check=False, capture_output=True,
        ).stdout.decode()
    except FileNotFoundError:
        logging.getLogger(__name__).exception("please install lshw: sudo apt install lshw")
    return out


def get_lspci() -> str:
    """Extract the information with lspci -k.

    Examples
    --------
    >>> from mendevi.measures.context import get_lspci
    >>> lspci_info = get_lspci()
    >>>

    """
    try:
        out = subprocess.run(
            ["lspci", "-k"], check=False, capture_output=True,
        ).stdout.decode()
    except FileNotFoundError:
        logging.getLogger(__name__).exception("please install lshw: sudo apt install lshw")
    return out


def full_context() -> dict:
    """Get the full context informations."""
    return {
        "ffmpeg_version": get_ffmpeg_version(),
        "librav1e_version": get_librav1e_version(),
        "libsvtav1_version": get_libsvtav1_version(),
        "libvpx_vp9_version": get_libvpx_vp9_version(),
        "libx265_version": get_libx265_version(),
        "lshw": get_lshw(),
        "lspci": get_lspci(),
        "pip_freeze": get_pip_freeze(),
        "vvc_version": get_vvc_version(),
        **get_platform(),
    }



"""Define the functions that enable values to be extracted from a select query."""

import functools
import numbers
import pathlib
import re
import typing

import numpy as np
import orjson

from mendevi.cst.profiles import PROFILES
from mendevi.database.serialize import binary_to_list, binary_to_tensor

JOIN: dict[str: dict[str, str]] = {  # join = JOIN[destination_table][source_table]
    "t_vid_video": {
        "t_dec_decode": "JOIN t_vid_video ON t_dec_decode.dec_vid_id = t_vid_video.vid_id",
        "t_enc_encode": "JOIN t_vid_video ON t_enc_encode.enc_dst_vid_id = t_vid_video.vid_id",
        "t_met_metric": "JOIN t_vid_video ON t_met_metric.met_dis_vid_id = t_vid_video.vid_id",
    },
    "t_enc_encode": {
        "t_dec_decode": (
            # rather "JOIN t_enc_encode ON t_dec_decode.dec_vid_id = t_enc_encode.enc_dst_vid_id"
            # this requets select only one line of t_enc_encode
            "JOIN (\n"
            "    SELECT t_enc_encode.enc_dst_vid_id, MIN(enc_id) AS enc_id_min\n"
            "    FROM t_enc_encode\n"
            "    GROUP BY enc_dst_vid_id\n"
            ") AS t_enc_encode_single_dst_vid_id "
            "ON t_dec_decode.dec_vid_id = t_enc_encode_single_dst_vid_id.enc_dst_vid_id\n"
            "JOIN t_enc_encode ON t_enc_encode.enc_id = t_enc_encode_single_dst_vid_id.enc_id_min"
        ),
    },
    "t_dec_decode": {
        "t_enc_encode": (
            "JOIN t_dec_decode ON t_enc_encode.enc_dst_vid_id = t_dec_decode.dec_vid_id"
        ),
    },
    "t_met_metric": {
        "t_vid_video": "LEFT JOIN t_met_metric ON t_vid_video.vid_id = t_met_metric.met_dis_vid_id",
        "t_enc_encode": (
            "LEFT JOIN t_met_metric ON t_enc_encode.enc_dst_vid_id = t_met_metric.met_dis_vid_id "
            "AND t_enc_encode.enc_src_vid_id = t_met_metric.met_ref_vid_id"
        ),
        "t_dec_decode": (
            "LEFT JOIN t_met_metric ON t_dec_decode.dec_vid_id = t_met_metric.met_dis_vid_id"
        ),
    },
    "t_env_environment": {
        "t_dec_decode": (
            "JOIN t_env_environment ON t_dec_decode.dec_env_id = t_env_environment.env_id"
        ),
        "t_enc_encode": (
            "JOIN t_env_environment ON t_enc_encode.enc_env_id = t_env_environment.env_id"
        ),
    },
    "t_act_activity": {
        "t_dec_decode": "JOIN t_act_activity ON t_dec_decode.dec_act_id = t_act_activity.act_id",
        "t_enc_encode": "JOIN t_act_activity ON t_enc_encode.enc_act_id = t_act_activity.act_id",
    },
    "t_idle": {
        "t_env_environment": (
            "JOIN t_act_activity AS t_idle "
            "ON t_env_environment.env_idle_act_id = t_idle.act_id"
        ),
    },
    "t_ref_video": {  # the reference video
        "t_enc_encode": (
            "JOIN t_vid_video AS t_ref_video "
            "ON t_enc_encode.enc_src_vid_id = t_ref_video.vid_id"
        ),
        "t_dec_decode": (  # to avoid redundency, we may define only atomic steps then link them
            "JOIN (\n"
            "    SELECT t_enc_encode.enc_dst_vid_id, MIN(enc_id) AS enc_id_min\n"
            "    FROM t_enc_encode\n"
            "    GROUP BY enc_dst_vid_id\n"
            ") AS t_enc_encode_single_dst_vid_id_from_dec "
            "ON t_dec_decode.dec_vid_id = t_enc_encode_single_dst_vid_id_from_dec.enc_dst_vid_id\n"
            "JOIN t_enc_encode AS t_enc_from_dec "
            "ON t_enc_from_dec.enc_id = t_enc_encode_single_dst_vid_id_from_dec.enc_id_min\n"
            "JOIN t_vid_video AS t_ref_video "
            "ON t_enc_from_dec.enc_src_vid_id = t_ref_video.vid_id"
        ),
        "t_met_metric": (
            "JOIN t_vid_video AS t_ref_video "
            "ON t_met_metric.met_ref_vid_id = t_ref_video.vid_id"
        ),
    },
    "t_dst_video": {  # the transcoded video
        "t_enc_encode": (
            "JOIN t_vid_video AS t_dst_video "
            "ON t_enc_encode.enc_dst_vid_id = t_dst_video.vid_id"
        ),
        "t_dec_decode": (
            "JOIN t_vid_video AS t_dst_video "
            "ON t_dec_decode.dec_vid_id = t_dst_video.vid_id"
        ),
        "t_met_metric": (
            "JOIN t_vid_video AS t_dst_video "
            "ON t_met_metric.met_dis_vid_id = t_dst_video.vid_id"
        ),
    },
}
# JOIN["t_ref_video"]["t_dec_decode"] = (  # and example of path
#     f"{JOIN['t_enc_encode']['t_dec_decode']}\n"
#     f"{JOIN['t_ref_video']['t_enc_encode']}"
# )


class SqlLinker:
    """Allow you to add an SQL query to an extractor."""

    def __init__(self, *select: str) -> None:
        """Initialise the linker.

        Parameters
        ----------
        select : args[str]
            The fields to be returned (juste after SELECT), with the optional alias.

        """
        assert all(isinstance(s, str) for s in select), select
        self.select: list[str] = sorted(set(select))

    @property
    def sql(self) -> str:
        """Write the sql request."""
        # find all possible junctions
        dst_tables = {s.split(".")[0] for s in self.select}
        joins: dict[str] = {}
        for src_table in {t for j in JOIN.values() for t in j}:
            join: set[str] = set()
            for dst_table in dst_tables - {src_table}:
                if dst_table not in JOIN:
                    break
                if src_table not in JOIN[dst_table]:
                    break
                join.add(JOIN[dst_table][src_table])
            else:
                joins[src_table] = join

        # put in form the queries
        queries: list[str] = []
        for src_table in sorted(
            joins,
            key=lambda t: (
                len(joins[t]),
                {
                    "t_vid_video": 0,
                    "t_act_activity": 1,
                    "t_met_metric": 2,
                    "t_enc_encode": 3,
                    "t_dec_decode": 4,
                    "t_env_environment": 5,
                }[t],
            ),
        ):  # priority no join
            select_str = f"SELECT {', '.join(self.select)}"
            if len(select_str) >= 80:
                select_str = f"SELECT\n    {',\n    '.join(self.select)}"
            table_str = f"FROM {src_table}"
            if (join_str := "\n".join(
                re.sub(" ON ", "\n    ON ", j) for j in sorted(joins[src_table])
            )):
                sql = f"{select_str}\n{table_str}\n{join_str}"
            else:
                sql = f"{select_str}\n{table_str}"
            queries.append(sql)
        return queries

    def __call__(self, func: callable) -> callable:
        """Decorate a function.

        Returns
        -------
        A decorated function with the select.
        The docstring of the decorated function is also modified
        to illustrate the minimal SQL query with an example.

        """
        # set attributes
        func.select = self.select

        # set doctrsing
        doc: list[str] = (func.__doc__ or "").split("\n")
        example = "\nor, alternativaly\n".join(
            (
                "\n"
                ".. code:: sql\n"
                "\n"
                f"    {'\n    '.join(sql.split('\n'))}"
                "\n"
            )
            for sql in self.sql
        )
        doc.insert(1, example)
        func.__doc__ = "\n".join(doc)

        return func


def verif(func: typing.Callable) -> typing.Callable:
    """Perform few verifications."""
    @functools.wraps(func)
    def decorated(raw: dict[str]) -> float:
        assert isinstance(raw, dict), raw.__class__.__name__
        for select in func.select:
            name = select.split(" ")[-1].split(".")[-1]
            assert name in raw, \
                f"correct the SQL query because {select} is missing (in {func.__name__})"
        return func(raw)
    return decorated


@verif
@SqlLinker("t_act_activity.act_duration")
def extract_act_duration(raw: dict[str]) -> float:
    """Return the video processing activity duration in seconds.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (act_duration := raw["act_duration"]) is None:
        return None
    assert isinstance(act_duration, numbers.Real), act_duration.__class__.__name__
    assert act_duration > 0.0, act_duration.__class__.__name__
    return float(act_duration)


@verif
@SqlLinker("t_dst_video.vid_codec")
def extract_codec(raw: dict[str]) -> str:
    """Return the codec name.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (codec := raw["vid_codec"]) is None:
        return None
    assert isinstance(codec, str), codec.__class__.__name__
    return str(codec)


@verif
@SqlLinker("t_act_activity.act_ps_dt", "t_act_activity.act_ps_core")
def extract_cores(raw: dict[str]) -> float:
    """Return the average cumulative utilisation rate of logical cores.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    act_ps_dt, act_ps_core = raw["act_ps_dt"], raw["act_ps_core"]
    if act_ps_dt is None or act_ps_core is None:
        return None
    act_ps_dt = binary_to_list(act_ps_dt)
    act_ps_core = binary_to_tensor(act_ps_core).sum(axis=1)
    integral = (act_ps_core * act_ps_dt).sum()  # act_ps_core is already the average on each dt
    average = integral / act_ps_dt.sum()
    return float(average) / 100.0  # normalisation


@verif
@SqlLinker("t_dec_decode.dec_cmd", "t_vid_video.vid_name")
def extract_decode_cmd(raw: dict[str]) -> str:
    """Return the ffmpeg command used for decoding.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (cmd := raw["enc_cmd"]) is None:
        return None
    vid = raw["vid_name"] or "vid.mp4"
    return cmd.replace("vid.mp4", vid)


@verif
@SqlLinker("t_enc_encode.enc_effort")
def extract_effort(raw: dict[str]) -> str:
    """Return the effort provided as a parameter to the encoder.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (enc_effort := raw["enc_effort"]) is None:
        return None
    assert isinstance(enc_effort, str), enc_effort.__class__.__name__
    return str(enc_effort)


@verif
@SqlLinker("t_enc_encode.enc_cmd")
def extract_encode_cmd(raw: dict[str]) -> str:
    """Return the ffmpeg command used for encoding.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (cmd := raw["enc_cmd"]) is None:
        return None
    return cmd


@verif
@SqlLinker("t_enc_encode.enc_encoder")
def extract_encoder(raw: dict[str]) -> str:
    """Return the name of the encoder.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (encoder := raw["enc_encoder"]) is None:
        return None
    assert isinstance(encoder, str), encoder.__class__.__name__
    return str(encoder)


@verif
@SqlLinker("t_vid_video.vid_frames")
def extract_frames(raw: dict[str]) -> list[dict]:
    """Return the metadata of each frame.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (frames := raw["vid_frames"]) is None:
        return None
    return orjson.loads(frames)


@verif
@SqlLinker("t_dst_video.vid_gamut")
def extract_gamut(raw: dict[str]) -> str:
    """Return the tristimulus primaries colors name.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (gamut := raw["vid_gamut"]) is None:
        return None
    assert isinstance(gamut, str), gamut.__class__.__name__
    return gamut


@verif
@SqlLinker("t_dst_video.vid_height")
def extract_height(raw: dict[str]) -> int:
    """Return the height of images in pixels.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (height := raw["vid_height"]) is None:
        return None
    return int(height)


@verif
@SqlLinker("t_env_environment.env_hostname")
def extract_hostname(raw: dict[str]) -> str:
    """Return the machine name.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (hostname := raw["env_hostname"]) is None:
        return None
    return str(hostname)


@verif
@SqlLinker("t_met_metric.met_lpips_alex", "t_met_metric.met_lpips_vgg")
def extract_lpips(raw: dict[str]) -> float:
    """Return the Learned Perceptual Image Patch Similarity (LPIPS) with alex.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (lpips := raw["met_lpips_vgg"] or raw["met_lpips_alex"]) is None:
        return None
    return float(np.nanmean(binary_to_list(lpips)))


@verif
@SqlLinker("t_met_metric.met_lpips_alex")
def extract_lpips_alex(raw: dict[str]) -> float:
    """Return the Learned Perceptual Image Patch Similarity (LPIPS) with alex.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (lpips := raw["met_lpips_alex"]) is None:
        return None
    return float(np.nanmean(binary_to_list(lpips)))


@verif
@SqlLinker("t_met_metric.met_lpips_vgg")
def extract_lpips_vgg(raw: dict[str]) -> float:
    """Return the Learned Perceptual Image Patch Similarity (LPIPS) with vgg.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (lpips := raw["met_lpips_vgg"]) is None:
        return None
    return float(np.nanmean(binary_to_list(lpips)))


@verif
@SqlLinker("t_enc_encode.enc_mode")
def extract_mode(raw: dict[str]) -> str:
    """Return the bitrate mode, constant (cbr) or variable (vbr).

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (mode := raw["enc_mode"]) is None:
        return None
    assert isinstance(mode, str), mode.__class__.__name__
    assert mode in {"cbr", "vbr"}, mode
    return mode


@SqlLinker(
    "t_act_activity.act_rapl_dt",
    "t_act_activity.act_rapl_power",
    "t_act_activity.act_wattmeter_dt",
    "t_act_activity.act_wattmeter_power",
)
def extract_powers(raw: dict[str]) -> tuple:
    """Return the interval duration and the average power in each intervals.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    assert isinstance(raw, dict), raw.__class__.__name__
    assert (
        ("act_wattmeter_dt" in raw and "act_wattmeter_power" in raw)
        or ("act_rapl_dt" in raw and "act_rapl_power" in raw)
    ), "Please correct the SQL query."
    act_dt, act_power = raw["act_wattmeter_dt"], raw["act_wattmeter_power"]
    if act_dt is not None and act_power is not None:  # if it comes from wattmeter
        act_dt, act_power = binary_to_list(act_dt), binary_to_list(act_power)
        act_power = 0.5 * (act_power[:-1] + act_power[1:])  # trapez method
    else:  # if it comes from rapl
        act_dt, act_power = raw["act_rapl_dt"], raw["act_rapl_power"]
        if act_dt is None or act_power is None:
            return None
        act_dt, act_power = binary_to_list(act_dt), binary_to_list(act_power)
    return act_dt, act_power


@verif
@SqlLinker("t_met_metric.met_psnr")
def extract_psnr(raw: dict[str]) -> float:
    """Return the Peak Signal to Noise Ratio (PSNR).

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (psnr := raw["met_psnr"]) is None:
        return None
    return float(np.nanmean(binary_to_list(psnr)))


@verif
@SqlLinker("t_enc_encode.enc_quality")
def extract_quality(raw: dict[str]) -> float:
    """Return the quality level passed to the encoder.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (enc_quality := raw["enc_quality"]) is None:
        return None
    assert isinstance(enc_quality, numbers.Real), enc_quality.__class__.__name__
    assert 0.0 <= enc_quality <= 1.0, enc_quality
    return float(enc_quality)


@verif
@SqlLinker("t_dst_video.vid_range")
def extract_range(raw: dict[str]) -> str:
    """Return the video encoding color range, 'tv' or 'pc'.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (range_ := raw["vid_range"]) is None:
        return None
    assert isinstance(range_, str), range_.__class__.__name__
    assert range_ in {"pc", "tv"}, range_
    return range_


@verif
@SqlLinker("t_ref_video.vid_name AS ref_vid_name")
def extract_reference_video_name(raw: dict[str]) -> str:
    """Return the input video basename.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (vid_name := raw["ref_vid_name"]) is None:
        return None
    assert isinstance(vid_name, str), vid_name.__class__.__name__
    return re.sub(fr"^reference_(\w+)_(?:{'|'.join(PROFILES)})\.\w+$", r"\1", vid_name)


@verif
@SqlLinker("t_vid_video.vid_rms_sobel")
def extract_rms_sobel(raw: dict[str]) -> str:
    """Return the spacial root mean square sobel gradient complexity.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (comp := raw["vid_rms_sobel"]) is None:
        return None
    return float(np.nanmean(binary_to_list(comp)))


@verif
@SqlLinker("t_vid_video.vid_rms_time_diff")
def extract_rms_time_diff(raw: dict[str]) -> str:
    """Return the temporal root means square time difference complexity.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (comp := raw["vid_rms_time_diff"]) is None:
        return None
    return float(np.nanmean(binary_to_list(comp)))


@verif
@SqlLinker("t_met_metric.met_ssim")
def extract_ssim(raw: dict[str]) -> float:
    """Return the Structural Similarity (SSIM).

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (ssim := raw["met_ssim"]) is None:
        return None
    return float(np.nanmean(binary_to_list(ssim)))


@verif
@SqlLinker("t_act_activity.act_ps_dt", "t_act_activity.act_ps_temp")
def extract_temp(raw: dict[str]) -> float:
    """Return the average temperature in C.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    act_ps_dt, act_ps_temp = raw["act_ps_dt"], raw["act_ps_temp"]
    if act_ps_dt is None or act_ps_temp is None:
        return None
    act_ps_dt = binary_to_list(act_ps_dt)
    act_ps_temp = orjson.loads(act_ps_temp)
    avg_temp = sum(np.asarray(tps, act_ps_dt.dtype) for tps in act_ps_temp.values())
    integral = (avg_temp * act_ps_dt).sum()
    average = integral / (act_ps_dt.sum() * len(act_ps_temp))
    return float(average)


@verif
@SqlLinker("t_enc_encode.enc_threads")
def extract_threads(raw: dict[str]) -> int:
    """Return the number of threads provided as a parameter to the encoder.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (threads := raw["enc_threads"]) is None:
        return None
    assert isinstance(threads, numbers.Integral), threads.__class__.__name__
    assert threads >= 1, threads.__class__.__name__
    return int(threads)


@verif
@SqlLinker("t_dst_video.vid_eotf")
def extract_transfer(raw: dict[str]) -> str:
    """Return the non-linear transfer function name.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (transfer := raw["vid_eotf"]) is None:
        return None
    assert isinstance(transfer, str), transfer.__class__.__name__
    return transfer


@verif
@SqlLinker("t_dst_video.vid_duration")
def extract_video_duration(raw: dict[str]) -> float:
    """Return the video duration in seconds.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (duration := raw["vid_duration"]) is None:
        return None
    assert isinstance(duration, numbers.Real), duration.__class__.__name__
    return float(duration)


@verif
@SqlLinker("t_vid_video.vid_id")
def extract_video_hash(raw: dict[str]) -> str:
    """Return the hexadecimal md5 video file checksum.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    return raw["vid_id"].hex()


@verif
@SqlLinker("t_vid_video.vid_name")
def extract_video_name(raw: dict[str]) -> str:
    """Return the video basename.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (vid_name := raw["vid_name"]) is None:
        return None
    assert isinstance(vid_name, str), vid_name.__class__.__name__
    return pathlib.Path(vid_name).name


@verif
@SqlLinker("t_dst_video.vid_size")
def extract_video_size(raw: dict[str]) -> int:
    """Return the total video file size in bytes.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (size := raw["vid_size"]) is None:
        return None
    assert isinstance(size, numbers.Integral), size.__class__.__name__
    return int(size)


@verif
@SqlLinker("t_met_metric.met_vmaf")
def extract_vmaf(raw: dict[str]) -> float:
    """Return the Video Multi-Method Assessment Fusion (VMAF).

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (vmaf := raw["met_vmaf"]) is None:
        return None
    return float(np.nanmean(binary_to_list(vmaf)))


@verif
@SqlLinker("t_dst_video.vid_width")
def extract_width(raw: dict[str]) -> int:
    """Return the width of images in pixels.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    """
    if (width := raw["vid_width"]) is None:
        return None
    return int(width)



"""Perform the video properties measures."""

import pathlib
import random

import click
from context_verbose import Printer

from mendevi.probe import probe_and_store
from mendevi.utils import compute_video_hash

from .parse import parse_videos_database


def _parse_args(prt: Printer, kwargs: dict) -> None:
    """Verification of the arguments."""
    kwargs["ref"] = kwargs.get("ref", ())
    assert isinstance(kwargs["ref"], tuple), kwargs["ref"]
    assert all(isinstance(r, str) for r in kwargs["ref"]), kwargs["ref"]
    kwargs["ref"] = [pathlib.Path(r).expanduser() for r in kwargs["ref"]]
    assert all(r.is_file() for r in kwargs["ref"]), kwargs["ref"]
    if kwargs["ref"]:
        prt.print(f"ref       : {', '.join(r.name for r in kwargs['ref'])}")
    for met in [
        "lpips_alex",
        "lpips_vgg",
        "psnr",
        "rms_sobel",
        "rms_time_diff",
        "ssim",
        "uvq",
        "vif",
        "vmaf",
    ]:
        assert met in kwargs, f"{met} not in {sorted(kwargs)}"
        assert isinstance(kwargs[met], bool), kwargs[met].__class__.__name__
        prt.print(f"{met:<10}: {'yes' if kwargs[met] else 'no'}")


@click.command()
@click.argument("videos", nargs=-1, type=click.Path())
@click.option("-d", "--database", type=click.Path(), help="The database path.")
@click.option(
    "-r", "--ref",
    type=click.Path(),
    multiple=True,
    help="The reference video for comparative metrics.",
)
@click.option("--lpips-alex/--no-lpips-alex", default=False, help="Compute the LPIPS-ALEX or not.")
@click.option("--lpips-vgg/--no-lpips-vgg", default=False, help="Compute the LPIPS-VGG or not.")
@click.option("--psnr/--no-psnr", default=True, help="Compute the PSNR or not.")
@click.option("--ssim/--no-ssim", default=True, help="Compute the SSIM or not.")
@click.option("--rms-sobel/--no-rms-sobel", default=True, help="Compute the C_sob or not.")
@click.option("--rms-time-diff/--no-rms-time-diff", default=True, help="Compute the C_td or not.")
@click.option("--uvq/--no-uvq", default=False, help="Compute the UVQ or not.")
@click.option("--vif/--no-vif", default=False, help="Compute the VIF or not.")
@click.option("--vmaf/--no-vmaf", default=False, help="Compute the VMAF or not.")
def main(videos: tuple[str], database: str | None = None, **kwargs: dict) -> None:
    """Measure the video properties.

    \b
    Parameters
    ----------
    videos : tuple[pathlike], optional
        All videos to be analysed. It can be a glob expression, a directory or a file path.
    database : pathlike, optional
        The path to the database where all measurements are stored.
        If a folder is provided, the database is created inside this folder.
    **kwargs: dict
        Please refer to the detailed arguments below.
    ref : tuple(pathlike), optional
        If provided, in addition to the original video ``o``,
        which has been transcoded to produce video ``x``,
        the metrics will also be calculated in relation to these reference videos ``ref``.
        Finaly comparatives metrics will be between ``o`` and ``x``, and also ``ref`` and ``x``.
    lpips_alex, lpips_vgg, psnr, ssim, vif, vmaf : boolean
        If True, compute the comparative video metric, otherwise, skip it.
    rms_sobel, rms_time_diff, uvq : boolean
        If True, compute the absolute uvq metric, overwise, skip it.

    """
    with Printer("Parse configuration...") as prt:
        videos, database = parse_videos_database(prt, videos, database)
        _parse_args(prt, kwargs)

    kwargs["ref"]: dict[pathlib.Path, bytes] = compute_video_hash(kwargs["ref"])

    # do the job
    random.shuffle(videos)  # heuristic to improve efficiency of multiple acces
    conn = None  # to avoid to many open/close connections (keep alive)
    for i, video in enumerate(sorted(kwargs["ref"]) + videos):
        with Printer(f"Probe {i+1}/{len(videos)+len(kwargs['ref'])}...", color="cyan") as prt:
            conn = probe_and_store(database, video, conn=conn, **kwargs)
            prt.print_time()
